"""
This module updates the qiprofile database clinical information
from the clinical Excel workbook file.
"""
from . import (xls, demographics, breast_pathology, sarcoma_pathology,
               chemotherapy, radiotherapy)


class ClinicalError(Exception):
    pass


def update(subject, filename):
    """
    Updates the subject clinical database content from the given workbook
    file.

    :param subject: the target qiprofile Subject to update
    :param filename: the input file location
    """
    wb = xls.load_workbook(filename)
    pathology_module = _pathology_module(subject.collection)
    for module in [demographics, pathology_module, chemotherapy, radiotherapy]:
        rows = module.read(wb, subject_number=subject.number)
        module.update(subject, rows)
    subject.save()


def _pathology_module(collection):
    """
    :param collection: the subject collection name
    :return: the pathology module
    :raise ClinicalError: if the collection does not have a pathology module
    """
    if collection.lower() == 'breast':
        return breast_pathology
    elif collection.lower() == 'sarcoma':
        return sarcoma_pathology
    else:
        raise ClinicalError("Pathology collection not supported: %s" %
                             collection)



import tempfile
import logging
from nipype.pipeline import engine as pe
from nipype.interfaces.utility import (IdentityInterface, Function)
from nipype.interfaces import fsl
from ..interfaces import (XNATUpload, MriVolCluster)
from .workflow_base import WorkflowBase
from qiutil.logging import logger

RESOURCE = 'mask'
"""The XNAT mask resource name."""

TIME_SERIES = 'scan_ts'
"""The XNAT scan time series resource name."""


def run(project, subject, session, scan, time_series, **opts):
    """
    Creates a :class:`qipipe.pipeline.mask.MaskWorkflow` and runs it
    on the given inputs.

    :param project: the project name
    :param subject: the input subject
    :param session: the input session
    :param scan: the input scan number
    :param time_series: the input 4D NIfTI time series to mask
    :param opts: additional :class:`MaskWorkflow` initialization parameters
    :return: the XNAT mask resource name
    """
    wf_gen = MaskWorkflow(project=project, **opts)

    return wf_gen.run(subject, session, scan, time_series)


class MaskWorkflow(WorkflowBase):
    """
    The MaskWorkflow class builds and executes the mask workflow.

    The workflow creates a mask to subtract extraneous tissue for a given
    input session 4D NIfTI time series. The new mask is uploaded to XNAT
    as a session resource named ``mask``.

    The mask workflow input is the `input_spec` node consisting of
    the following input fields:

     - subject: the XNAT subject name

     - session: the XNAT session name

     - scan: the XNAT scan number

     - time_series: the 4D NIfTI series image file

    The mask workflow output is the `output_spec` node consisting of the
    following output field:

    - `mask`: the mask file

    The optional workflow configuration file can contain the following
    sections:

    - ``fsl.MriVolCluster``: the
        :class:`qipipe.interfaces.mri_volcluster.MriVolCluster`
        interface options
    """

    def __init__(self, **kwargs):
        """
        If the optional configuration file is specified, then the workflow
        settings in that file override the default settings.

        :param kwargs: the :class:`qipipe.pipeline.workflow_base.WorkflowBase`
            initializer keyword arguments
        """
        super(MaskWorkflow, self).__init__(logger=logger(__name__), **kwargs)

        self.workflow = self._create_workflow()
        """The mask creation workflow."""

    def run(self, subject, session, scan, time_series):
        """
        Runs the mask workflow on the scan NIfTI files for the given
        time series.

        :param subject: the input subject
        :param session: the input session
        :param scan: the input scan number
        :param time_series: the input 3D NIfTI time series to mask
        :return: the mask XNAT resource name
        """
        self.logger.debug("Creating the mask for the %s %s scan %d time series"
                           " %s..." % (subject, session, scan, time_series))
        self.set_inputs(subject, session, scan, time_series)
        # Execute the workflow.
        self._run_workflow(self.workflow)
        self.logger.debug("Created the %s %s scan %s time series %s mask XNAT"
                           " resource %s." %
                           (subject, session, scan, time_series, RESOURCE))

        # Return the mask XNAT resource name.
        return RESOURCE

    def set_inputs(self, subject, session, scan, time_series):
        # Set the inputs.
        input_spec = self.workflow.get_node('input_spec')
        input_spec.inputs.subject = subject
        input_spec.inputs.session = session
        input_spec.inputs.scan = scan
        input_spec.inputs.time_series = time_series

    def _create_workflow(self):
        """
        Creates the mask workflow.

        :return: the Workflow object
        """
        self.logger.debug('Creating the mask reusable workflow...')
        workflow = pe.Workflow(name='mask', base_dir=self.base_dir)

        # The workflow input.
        in_fields = ['subject', 'session', 'scan', 'time_series']
        input_spec = pe.Node(IdentityInterface(fields=in_fields),
                             name='input_spec')

        # Get a mean image from the DCE data.
        dce_mean = pe.Node(fsl.MeanImage(), name='dce_mean')
        workflow.connect(input_spec, 'time_series', dce_mean, 'in_file')

        # Find the center of gravity from the mean image.
        find_cog = pe.Node(fsl.ImageStats(), name='find_cog')
        find_cog.inputs.op_string = '-C'
        workflow.connect(dce_mean, 'out_file', find_cog, 'in_file')

        # Zero everything posterior to the center of gravity on mean image.
        crop_back = pe.Node(fsl.ImageMaths(), name='crop_back')
        workflow.connect(dce_mean, 'out_file', crop_back, 'in_file')
        workflow.connect(find_cog, ('out_stat', _gen_crop_op_string),
                         crop_back, 'op_string')

        # The cluster options.
        # Find large clusters of empty space on the cropped image.
        cluster_mask = pe.Node(MriVolCluster(), name='cluster_mask')
        workflow.connect(crop_back, 'out_file', cluster_mask, 'in_file')

        # Convert the cluster labels to a binary mask.
        binarize = pe.Node(fsl.BinaryMaths(), name='binarize')
        binarize.inputs.operation = 'min'
        binarize.inputs.operand_value = 1
        workflow.connect(cluster_mask, 'out_cluster_file', binarize, 'in_file')

        # Invert the binary mask.
        inv_mask_xfc = fsl.maths.MathsCommand(args='-sub 1 -mul -1',
                                              out_file='mask.nii.gz')
        inv_mask = pe.Node(inv_mask_xfc, name='inv_mask')
        workflow.connect(binarize, 'out_file', inv_mask, 'in_file')

        # Upload the mask to XNAT.
        upload_mask_xfc = XNATUpload(project=self.project, resource=RESOURCE,
                                     modality='MR')
        upload_mask = pe.Node(upload_mask_xfc, name='upload_mask')
        workflow.connect(input_spec, 'subject', upload_mask, 'subject')
        workflow.connect(input_spec, 'session', upload_mask, 'session')
        workflow.connect(input_spec, 'scan', upload_mask, 'scan')
        workflow.connect(inv_mask, 'out_file', upload_mask, 'in_files')

        # The output is the mask file.
        output_spec = pe.Node(IdentityInterface(fields=['mask']),
                                                name='output_spec')
        workflow.connect(inv_mask, 'out_file', output_spec, 'mask')

        self._configure_nodes(workflow)

        self.logger.debug("Created the %s workflow." % workflow.name)
        # If debug is set, then diagram the workflow graph.
        if self.logger.level <= logging.DEBUG:
            self.depict_workflow(workflow)

        return workflow


def _gen_crop_op_string(cog):
    """
    :param cog: the center of gravity
    :return: the crop -roi option
    """
    return "-roi 0 -1 %d -1 0 -1 0 -1" % cog[1]


def _crop_posterior(image, cog):
    from nipype.interfaces import fsl

    crop_back = fsl.ImageMaths()
    crop_back.inputs.op_string = '-roi 0 -1 %d -1 0 -1 0 -1' % cog[1]
    crop_back.inputs.in_file = image
    return crop_back.run().outputs.out_file



"""
This module updates the qiprofile database Subject drug dosage
information from a Chemotherapy Excel worksheet.
"""
import datetime
from qiprofile_rest_client.model.clinical import (Treatment, Dosage)
from .xls import Worksheet
from . import parse


class DosageError(Exception):
    pass


class DosageWorksheet(Worksheet):
    """The dosage worksheet facade."""
    
    def __init__(self, workbook, sheet, agent_class, **opts):
        """
        :param workbook: the :class:`qipipe.qiprofile.xls.Workbook` object
        :param sheet: the sheet name
        :param agent_class: the agent class
        :param opts: the additional :class:`qipipe.qiprofile.xls.Worksheet`
            initializer options
        """
        # Delegate to the superclass with the Treatment, Dosage and agent
        # classes.
        super(DosageWorksheet, self).__init__(
            workbook, sheet, Treatment, Dosage, agent_class, **opts
        )


class DosageUpdate(object):
    """The dosage update abstract class."""
    
    DEFAULTS = dict(duration=1)
    """The default duration is 1 day."""
    
    def __init__(self, subject, agent_class, **defaults):
        """
        :param subject: the ``Subject`` Mongo Engine database object
            to update
        :param agent_class: the dosage agent class
        :param defaults: the {attribute: value} row defaults
        """
        self._subject = subject
        for attr, val in DosageUpdate.DEFAULTS.iteritems():
            if attr not in defaults:
                defaults[attr] = val
        self._agent_class = agent_class
        self._defaults = defaults
    
    def update(self, rows):
        """
        Updates the subject data object from the given dosage XLS rows.

        :param rows: the input dosage
            :meth:`qipipe.qiprofile.xls.Worksheet.read` rows list 
        """
        for row in rows:
            self._update_for(row)

    def _update_for(self, row):
        """
        :param row: the input dosage
            :meth:`qipipe.qiprofile.xls.Worksheet.read` row
        """
        # Apply the defaults.
        for attr, val in self._defaults.iteritems():
            if getattr(row, attr) == None:
                setattr(row, attr, val)

        # The treatment object for the input treatment type.
        trt = self._treatment_for(row.treatment_type)
        # Extend the treatment span, if necessary.
        if not trt.start_date or trt.start_date > row.start_date:
            trt.start_date = row.start_date
        delta = datetime.timedelta(row.duration)
        end_date = row.start_date + delta
        if not trt.end_date or trt.end_date < end_date:
            trt.end_date = end_date

        # If there is no amount, then we can only store the
        # treatment without dosages.
        if not row.amount:
            return
        # Find or make the target dosage object.
        dosage = self._dosage_for(trt, row)
        # Collect the update attributes.
        attrs = (attr for attr in Dosage._fields if attr in row)
        # Update the target dosage database object.
        for attr in attrs:
            setattr(dosage, attr, row[attr])

    def _treatment_for(self, treatment_type):
        """
        :param treatment_type: the treatment type, e.g. ``adjuvant``
        :return: the existing or new treatment object
        """
        # Find the matching treatment, if it exists.
        trt_iter = (trt for trt in self._subject.treatments
                    if trt.treatment_type == treatment_type)
        target = next(trt_iter, None)
        if not target:
            target = Treatment(treatment_type=treatment_type)
            self._subject.treatments.append(target)

        # Return the target treatment.
        return target

    def _dosage_for(self, treatment, row):
        """
        :param treatment: the target treatment
        :param row: the input row
        :return: the dosage database object which matches the start
            date and agent attribute, or a new dosage database 
            object if there is no match
        :raise DosageError: if the key does not include the *start_date*
        """
        # Find the matching treatment dosage by start date and agent
        # content.
        dosage_iter = (dosage for dosage in treatment.dosages
                       if self._is_dosage_match(dosage, row))
        target = next(dosage_iter, None)
        # If no match, then make a new dosage database object.
        if not target:
            # The dosages are in date order. Insert the new dosage
            # directly before other dosages with a later date, or
            # the end of the list is there is no later dosage.
            index_iter = (i for i, dosage in enumerate(treatment.dosages)
                          if dosage.start_date > row.start_date)
            # The dosage insertion point.
            index = next(index_iter, len(treatment.dosages))
            # The new agent {attribute: value} content.
            agent_content = {attr: getattr(row, attr)
                             for attr in self._agent_class._fields
                             if hasattr(row, attr)}
            # The new agent object.
            agent = self._agent_class(**agent_content)
            # The new dosage object.
            target = Dosage(agent=agent, start_date=row.start_date)
            # Add the new dosage.
            treatment.dosages.insert(index, target)

        return target

    def _is_dosage_match(self, dosage, row):
        return (dosage.start_date == row.start_date
                and self._is_agent_match(dosage.agent, row))

    def _is_agent_match(self, agent, row):
        return all(getattr(agent, attr) == getattr(row, attr)
                   for attr in agent.__class__._fields
                   if hasattr(row, attr))
        


"""The top-level Quantitative Imaging Pipeline module."""
import os

__version__ = '5.6.3'
"""
The one-based major.minor.patch version based on the
`Fast and Loose Versioning <https://gist.github.com/FredLoney/6d946112e0b0f2fc4b57>`_
scheme. Minor and patch version numbers begin at 1.
"""

CONF_DIR = os.path.join(os.path.dirname(__file__), 'conf')
"""The configuration directory."""



import re
import openpyxl
from contextlib import contextmanager
from datetime import datetime
from bunch import bunchify
import inflection
from qiutil.logging import logger
from . import parse


class XLSError(Exception):
    pass


def load_workbook(filename):
    """
    :param filename: the XLS workbook file location
    :return the read-only ``openpyxl`` workbook object
    """
    return openpyxl.load_workbook(filename, read_only=True)


class Worksheet(object):

    PAREN_SUFFIX_REGEX = re.compile('\s*\(.*\)$')
    """Matches a parenthesis suffix."""

    NONWORD_REGEX = re.compile('\W+')
    """Matches a non-word string."""

    """
    Excel worksheet facade. The worksheet rows are required to consist of
    a column heading row followed by the data rows.
    """

    def __init__(self, workbook, sheet, *classes, **opts):
        """
        :param workbook: the openpyxl workbook
        :param sheet: the sheet name
        :param classes: the classes which have spreadsheet attributes
        :param opts: the following options:
        :option parsers: the non-standard
            {attribute: value parser function} parsers dictionary
        :param column_attributes: the non-standard
            {column name: attribute} dictionary
        :raise XLSError: if there are unparsed columns
        """
        self.worksheet = workbook[sheet]
        """The wrapped openpyxl worksheet."""

        # The default parsers.
        parsers = parse.default_parsers(*classes)
        # Every worksheet has a subject number field.
        parsers.update(subject_number=int)
        # Add the special parsers.
        parsers_opt = opts.get('parsers', None)
        if parsers_opt:
            parsers.update(parsers_opt)
        self._parsers = parsers
        """The {attribute: value parser function} dictionary,"""

        # The first row is the headings.
        hdg_row = next(self.worksheet.iter_rows())
        # The column headings stop when the first blank cell is reached.
        headings = []
        for cell in hdg_row:
            if not cell.value:
                break
            headings.append(cell.value)
        # The non-standard {column name: attribute} dictionary.
        col_attrs = opts.get('column_attributes', {})
        # The attributes in the same order as the headings. The default
        # column attribute is generated by the _attributize function.
        self.attributes = [col_attrs.get(hdg, self._attributize(hdg))
                           for hdg in headings]
        """The row attributes in column order."""

        # There must be a parser for each attribute.
        unparsed = [attr for attr in self.attributes
                    if not attr in self._parsers]
        if unparsed:
            raise XLSError("%s columns are missing a parser: %s" %
                           (sheet, unparsed))

    def read(self, **condition):
        """
        Returns a row generator, where each row is a {attribute: value}
        bunch. This generator yields each row which satisfies the
        following condition:
        1. the row is non-empty, i.e. has at least one cell value, and
        2. if this reader has a filter, then the row satisfies the
           filter condition

        :param condition: the {attribute: value} row filter condition
        :return: the row {attribute: value} bunch generator
        """
        reader = Reader(self.worksheet, self.attributes, **condition)

        return (self._parse_row(row) for row in reader.read())

    def _parse_row(self, row):
        """
        Extracts and parses the row values.

        :param row: the XLS row to parse
        :return: the row {attribute: value} bunch
        """
        # The {attribute: value} dictionary.
        attr_val_dict = {}
        for i, cell in enumerate(row):
            # Stop when we are past the attributes.
            if i >= len(self.attributes):
                break
            # The attribute key.
            attr = self.attributes[i]
            # The value parser.
            parser = self._parsers[attr]
            # The attribute value.
            if cell.value == None or cell.value == '':
                val = None
            else:
                val = parser(cell.value)            
            # Add the attribute value entry.
            attr_val_dict[attr] = val

        # Return the {attribute: value} bunch.
        return bunchify(attr_val_dict)

    def _attributize(self, s):
        """
        Converts the given string to an underscore attribute.
        A trailing parenthesized word is not included.

        Examples::

            >> reader._attributize('Start Date')
            start_date
            >> reader._attributize('Amount (mg/kg)')
            amount

        :param s: the string to convert
        :return: the underscore attribute name
        """
        stripped = Worksheet.PAREN_SUFFIX_REGEX.sub('', s)
        word = Worksheet.NONWORD_REGEX.sub('_', stripped)

        return inflection.underscore(word)


class Reader(object):
    """Reads an Excel worksheet."""

    def __init__(self, worksheet, attributes, **condition):
        """
        :param worksheet: the :attr:`worksheet` object
        :param conditional: the optional {attribute: value} row filter
            condition
        """
        self.worksheet = worksheet
        """The wrapped openpyxl worksheet."""
        
        # Validate the condition.
        for key in condition:
            if not key in attributes:
                raise XLSError("The condition key %s is not in the attribute"
                                " list" % key)
        # Capture the attribute range for iteration reuse.
        self._attr_range = range(0, len(attributes))
        # Convert the {attribute: value} dictionary to a
        # {column index: value} dictonary.
        self._filter = {attributes.index(attr): val
                        for attr, val in condition.iteritems()}

    def read(self):
        """
        Returns a row generator, where each row is a {attribute: value}
        bunch. This generator yields each row which satisfies the
        following condition:
        1. the row is non-empty, i.e. has at least one cell value, and
        2. if this reader has a filter, then the row satisfies the
           filter condition

        :return: the filtered ``openpyxl`` row iterator
        """
        # Start reading after the header row.
        row_iter = self.worksheet.iter_rows(row_offset=1)
        return (row for row in row_iter if self._include_row(row))

    def _include_row(self, row):
        """
        :param row: the row to check
        :return: whether to include the row in the iteration
        """
        return self._is_row_nonempty(row) and self._filter_row(row)

    def _is_row_nonempty(self, row):
        """
        :param row: the row to check
        :return: whether the row has at least one cell value
        """
        return any(row[i].value != None for i in self._attr_range)

    def _filter_row(self, row):
        """
        :param row: the row to check
        :return: whether the row satisfies the filter condition
        """
        return all(row[i].value == val for i, val in self._filter.iteritems())        



import re
import functools
import six
import mongoengine
from qiutil import functions

TRAILING_NUM_REGEX = re.compile("(\d+)$")
"""A regular expression to extract the trailing number from a string."""

TRUE_REGEX = re.compile("(T(rue)?|Pos(itive)?|Present|Y(es)?)$", re.IGNORECASE)
"""
The valid True string representations are a case-insensitive match
for ``T(rue)?``, ``Pos(itive)?``, ``Present`` or ``Y(es)?``.
"""

FALSE_REGEX = re.compile("(F(alse)?|Neg(ative)?|Absent|N(o)?)$", re.IGNORECASE)
"""
The valid False string representations are a case-insensitive match
for ``F(alse)?``, ``Neg(ative)?``, ``Absent`` or ``N(o)?``.
"""

COMMA_DELIM_REGEX = re.compile(",\w*")
"""Match a comma with optional white space."""

TYPE_PARSERS = {
    mongoengine.fields.StringField: str,
    mongoengine.fields.IntField: int,
    mongoengine.fields.FloatField: float,
    # openpyxls dates are already the correct type.
    mongoengine.fields.DateTimeField: lambda v: v,
    # Wrap the functions below with a lambda as a convenience to allow
    # a forward reference to the parse functions defined below.
    mongoengine.fields.BooleanField: lambda v: parse_boolean(v),
    mongoengine.fields.ListField: lambda v: parse_list_string(v)
}
"""
The following type cast conversion parsers:
* string field => ``str``
* integer field => ``int``
* float field => ``float``
* boolean field => :meth:`parse_boolean`
* list field => :meth:`parse_list_string`
"""


class ParseError(Exception):
    pass


def parse_trailing_number(s):
    """
    :param s: the input string
    :return: the trailing number in the string
    :raise ParseError: if the input string does not have a trailing
        number
    """
    match = TRAILING_NUM_REGEX.search(s)
    if not match:
        raise ParseError("The input string does not have a trailing number:"
                         " %s" % s)
    
    return int(match.group(1))


def extract_trailing_number(value):
    """
    Returns the integer at the end of the given input value, as follows:
    * If the input value is an integer, then the result is the input
      value.
    * Otherwise, if the input value has a string type, then the result
      is the trailing digits converted to an integer.
    * Any other value is an error.
    :param the input integer or string
    :return: the trailing integer
    :raise ParseError: if the input value type is not int or a string type
    """
    if isinstance(value, int):
        return value
    elif isinstance(value, six.string_types):
        return parse_trailing_number(value)
    else:
        raise ParseError("Cannot extract a trailing number from the value"
                         " %s of type %s" % (value, value.__class__))


def parse_list_string(value):
    """
    Converts a comma-separated list input string to a list, e.g.:
    
    >> from qipipe.qiprofile.parse import parse_list_string
    >> parse_list_string('White, Asian')
    ['White', 'Asian']

    :param value: the input value
    :return: the value converted to a list
    """
    if isinstance(value, six.string_types):
        return [word.strip() for word in COMMA_DELIM_REGEX.split(value)]
    else:
        return [value]


def parse_boolean(value):
    """
    Parses the input value as follows:
    * If the input value is already a boolean, then return the value
    * If the input is None or the empty string, then return None
    * Otherwise, if the input is a string which matches
        :const:`TRUE_REGEX`, then return True
    * Otherwise, if the input is a string which matches
        :const:`FALSE_REGEX`, then return False 
    * Any other value is an error.

    :param value: the input value
    :return: the value as a boolean
    :raise ParseError: if the value cannot be converted
    """
    if isinstance(value, bool):
        return value
    elif not isinstance(value, six.string_types):
        raise ParseError("The input type cannot be converted to a boolean:"
                         " %s (%s)" % (value, value.__class__))
    elif not value:
        return None
    elif TRUE_REGEX.match(value):
        return True
    elif FALSE_REGEX.match(value):
        return False
    else:
        raise ParseError("The input value cannot be converted a boolean value:"
                         " %s" % value)


def default_parsers(*classes):
    """
    Associates the data model class fields to a parse function composed
    as follows:
    
    * The type cast function in :const:`TYPE_PARSERS`, if present
    * The controlled value lookup, if the field has controlled values
    
    :param classes: the data model classes
    :return: the {attribute: function} dictionary
    """
    parsers = {}
    for klass in classes:
        parsers.update(_default_parsers(klass))

    return parsers


def _default_parsers(klass):
    """
    :param klass: the data model class
    :return: the {attribute: function} dictionary
    """
    # The (attribute, parser or None) tuple generator.
    parsers = {}
    for attr, field in klass._fields.iteritems():
        parser = _default_parser(field)
        if parser:
            parsers[attr] = parser
    
    return parsers


def _default_parser(field):
    type_parser = _type_value_parser(field)
    cv_parser = _controlled_value_parser(field)
    # Compose CV look-up with type casting, allowing for
    # the possibility that one or both might be missing.
    parsers = [p for p in [cv_parser, type_parser] if p]

    return functions.compose(*parsers) if parsers else None


def _controlled_value_parser(field):
    """
    Associates the field to :meth:`controlled_value_for`
    if the field has controlled values.
    
    :param field: the data model field
    :return: the parser function, or None if this field does not have
        controlled values
    """
    if _has_controlled_values(field):
        return functools.partial(_controlled_value_for, field=field)


def _type_value_parser(field):
    """
    Returns the type cast conversion parser, as follows:
    * If the field has controlled value tuples, then convert the
      value to a string for the CV lookup
    * Otherwise, if the field has a :const:`TYPE_PARSERS` association,
      then use the associated parser
    * Ohterwise, no conversion is performed
    
    :param field: the data model field
    :return: the type cast conversion parser function
    """
    if field.choices and isinstance(field.choices[0], tuple):
        return str
    for field_type, parser in TYPE_PARSERS.iteritems():
        if isinstance(field, field_type):
            return parser


def _has_controlled_values(field):
    """
    :param field: the data model field
    :return: whether the field has controlled values
    """
    if isinstance(field, mongoengine.fields.ListField) and field.field:
        return _has_controlled_values(field.field)
    else:
        return field.choices != None


def _controlled_value_for(value, field):
    """
    Returns the controlled value which matches the given input value.
    The match is case-insensitive for strings.

    :param value: the input value
    :param field: the data model field object
    :return the matching controlled value
    :raise ParseError: if there is no match
    """
    # Recurse into a list.
    if isinstance(value, list):
        return [_controlled_value_for(v, field.field) for v in value]
    # The matching field choice.
    choice = _match_choice(value, field.choices)
    if choice == None:
        raise ParseError("The input %s value %s does not match one of"
                        " the supported field choices %s" %
                        (field, value, field.choices))

    # Return the controlled value specified by the choice.
    return _choice_controlled_value(choice)


def _match_choice(value, choices):
    for choice in choices:
        if _is_choice_match(value, choice):
            return choice


def _is_choice_match(value, choice):
    if isinstance(choice, tuple):
        return any((_is_choice_match(value, c) for c in choice))
    elif isinstance(choice, six.string_types):
        return str(value).lower() == choice.lower()
    else:
        return value == choice


def _choice_controlled_value(choice):
    return choice[0] if isinstance(choice, tuple) else choice



"""
This module contains the OHSU-specific image collections.

The following OHSU QIN scan numbers are captured:
    * 1: T1
    * 2: T2
    * 4: DWI
    * 6: PD
These scans have DICOM files specified by the
:attr:`qipipe.staging.collection.Collection.patterns`
:attr:`qipipe.staging.collection.Patterns.scan`
:attr:`qipipe.staging.collection.ScanPatterns.dicom`
attribute. The T1 scan has ROI files as well, specified by the
:attr:`qipipe.staging.collection.ScanPatterns.roi`
:attr:`qipipe.staging.collection.ROIPatterns.glob` and
:attr:`qipipe.staging.collection.ROIPatterns.regex` attributes
"""

import re
from .collection import (Collection, ScanPatterns, ROIPatterns)
from . import collections
from .staging_error import StagingError

MULTI_VOLUME_SCAN_NUMBERS = [1]
"""Only T1 scans can have more than one volume."""

BREAST_SUBJECT_REGEX = re.compile('BreastChemo(\d+)')
"""The Breast subject directory match pattern."""

SARCOMA_SUBJECT_REGEX = re.compile('Subj_(\d+)')
"""The Sarcoma subject directory match pattern."""

SESSION_REGEX_PAT = """
    (?:             # Don't capture the prefix
     [vV]isit       # The Visit or visit prefix form
     _?             # with an optional underscore delimiter
     |              # ...or...
     %s\d+_?V       # The alternate prefix form, beginning with
                    # a leading collection abbreviation
                    # substituted into the pattern below
    )               # End of the prefix
    (\d+)$          # The visit number
"""
"""
The session directory match pattern. This pattern must be specialized
for each collection by replacing the %s place-holder with a string.
"""

BREAST_SESSION_REGEX = re.compile(SESSION_REGEX_PAT % 'BC?', re.VERBOSE)
"""
The Sarcoma session directory match pattern. The variations
``Visit_3``, ``Visit3``, ``visit3``, ``BC4V3``, ``BC4_V3`` and ``B4V3``
all match Breast Session03.
"""

SARCOMA_SESSION_REGEX = re.compile(SESSION_REGEX_PAT % 'S', re.VERBOSE)
"""
The Sarcoma session directory match pattern. The variations
``Visit_3``, ``Visit3``, ``visit3`` ``S4V3``, and ``S4_V3`` all match
Sarcoma Session03.
"""

T1_PAT = '*concat*/*'
"""The T1 DICOM file match pattern."""

BREAST_T2_PAT = '*sorted/2_tirm_tra_bilat/*'
"""The Breast T2 DICOM file match pattern."""

SARCOMA_T2_PAT = '*T2*/*'
"""The Sarcoma T2 DICOM file match pattern."""

BREAST_DWI_PAT = '*sorted/*Diffusion/*'
"""The Breast DWI DICOM file match pattern."""

SARCOMA_DWI_PAT = '*Diffusion/*'
"""The Sarcoma DWI DICOM file match pattern."""

BREAST_PD_PAT = '*sorted/*PD*/*'
"""The Breast pseudo-proton density DICOM file match pattern."""

BREAST_ROI_PAT = 'processing/R10_0.[456]*/slice*/*.bqf'
"""
The Breast ROI glob filter. The ``.bqf`` ROI files are in the
following session subdirectory:

    processing/<R10 directory>/slice<slice index>/
"""

BREAST_ROI_REGEX = re.compile("""
    processing/                 # The visit processing subdirectory
    R10_0\.[456]                # The R10 series subdirectory
     (_L                        # The optional lesion modifier
      (?P<lesion>\d+)           # The lesion number
     )?                         # End of the lesion modifier
     /                          # End of the R10 subdirectory 
    slice                       # The slice subdirectory
     (?P<slice_sequence_number>\d+)       # The slice index
     /                          # End of the slice subdirectory
    (?P<fname>                  # The ROI file base name
     .*\.bqf                    # The ROI file extension
    )                           # End of the ROI file name
""", re.VERBOSE)
"""
The Breast ROI .bqf ROI file match pattern.
"""

SARCOMA_ROI_PAT = 'results/ROI_ave*/taui_d001/slice*/*.bqf'
"""
The Sarcoma ROI glob filter. The ``.bqf`` ROI files are in the
session subdirectory:

    results/<ROI directory>/slice<slice index>/
"""

SARCOMA_ROI_REGEX = re.compile("""
    results/                    # The visit processing subdirectory
    ROI_ave(rage)?/             # The ROI subdirectory
    taui_d001/                  # An intermediate sudirectory
    slice(?P<slice_sequence_number>\d+)/  # The slice subdirectory
    (?P<fname>.*\.bqf)          # The ROI file base name
""", re.VERBOSE)
"""
The Sarcoma ROI .bqf ROI file match pattern.

:Note: The Sarcoma ROI directories are inconsistently named, with several
    alternatives and duplicates.

    TODO - clarify which of the Sarcoma ROI naming variations should be used.

:Note: There are no apparent lesion number indicators in the Sarcoma ROI
    input.

    TODO - confirm that there is no Sarcoma lesion indicator.
"""

VOLUME_TAG = 'AcquisitionNumber'
"""
The DICOM tag which identifies the volume.
The OHSU QIN collections are unusual in that the DICOM images which
comprise a 3D volume have the same DICOM Series Number and Acquisition
Number tag. The series numbers are consecutive, non-sequential integers,
e.g. 9, 11, 13, ..., whereas the acquisition numbers are consecutive,
sequential integers starting at 1. The Acquisition Number tag is
selected as the volume number identifier.
"""


class BreastCollection(Collection):
    """The OHSU AIRC Breast collection."""

    def __init__(self):
        roi = ROIPatterns(glob=BREAST_ROI_PAT, regex=BREAST_ROI_REGEX)
        t1 = ScanPatterns(dicom=T1_PAT, roi=roi)
        t2 = ScanPatterns(dicom=BREAST_T2_PAT)
        dwi = ScanPatterns(dicom=BREAST_DWI_PAT)
        pd = ScanPatterns(dicom=BREAST_PD_PAT)
        scan = {1: t1, 2: t2, 4: dwi, 6: pd}
        opts = dict(subject=BREAST_SUBJECT_REGEX,
                    session=BREAST_SESSION_REGEX,
                    scan=scan, volume=VOLUME_TAG)
        super(BreastCollection, self).__init__('Breast', **opts)


class SarcomaCollection(Collection):
    """The OHSU AIRC Sarcoma collection."""

    def __init__(self):
        roi = ROIPatterns(glob=SARCOMA_ROI_PAT, regex=SARCOMA_ROI_REGEX)
        t1 = ScanPatterns(dicom=T1_PAT, roi=roi)
        t2 = ScanPatterns(dicom=SARCOMA_T2_PAT)
        dwi = ScanPatterns(dicom=SARCOMA_DWI_PAT)
        scan = {1: t1, 2: t2, 4: dwi}
        opts = dict(subject=SARCOMA_SUBJECT_REGEX,
                    session=SARCOMA_SESSION_REGEX,
                    scan=scan, volume=VOLUME_TAG)
        super(SarcomaCollection, self).__init__('Sarcoma', **opts)


# Create the OHSU QIN collections.
collections.add(BreastCollection(), SarcomaCollection())




import re
from nipype.interfaces.base import (traits, BaseInterfaceInputSpec,
                                    BaseInterface)
from ..helpers import qiprofile


class UpdateQIProfileInputSpec(BaseInterfaceInputSpec):
    project = traits.Str(mandatory=True, desc='The XNAT project id')

    subject = traits.Str(mandatory=True, desc='The XNAT subject name')

    session = traits.Str(mandatory=True, desc='The XNAT session name')


class UpdateQIProfile(BaseInterface):
    """
    The ``UpdateQIProfile`` Nipype interface updates the Imaging Profile
    database.
    """

    input_spec = UpdateQIProfileInputSpec

    def _run_interface(self, runtime):
        qiprofile.update_session(self.inputs.project, self.inputs.subject,
                               self.inputs.session)
        return runtime



"""
TCIA CTP preparation utilities.
"""

import sys
import os
import re
from .ctp_config import ctp_collection_for_name
from qiutil.logging import logger
from .staging_error import StagingError

# OHSU - this is OHSU-specific.
# TODO - make the format a config item.
PROP_FMT = 'QIN-%s-OHSU.ID-LOOKUP.properties'
"""
The format for the Patient ID map file name specified by CTP.
"""


def map_ctp(collection, *subjects, **opts):
    """
    Creates the TCIA patient id map. The map is written to a
    property file in the destination directory. The property file
    name is given by :meth:`property_filename`.
    
    :param collection: the image collection
    :param subjects: the subject names
    :param opts: the following keyword option:
    :keyword dest: the destination directory
    :return: the subject map file path
    """
    # Make the CTP id map.
    ctp_map = CTPPatientIdMap()
    ctp_map.add_subjects(collection, *subjects)
    # Write the id map property file.
    dest = opts.get('dest')
    if dest:
        dest = os.path.abspath(dest)
        if not os.path.exists(dest):
            os.makedirs(dest)
    else:
        dest = os.getcwd()
    logger(__name__).debug("Creating the TCIA subject map in %s..." % dest)
    out_file = os.path.join(dest, property_filename(collection))
    output = open(out_file, 'w')
    ctp_map.write(output)
    output.close()
    logger(__name__).debug("Created the TCIA subject map %s." % out_file)
    
    return out_file


def property_filename(collection):
    """
    Returns the CTP id map property file name for the given collection.
    The Sarcoma collection is capitalized in the file name, Breast is not.
    """
    if collection == 'Sarcoma':
        return PROP_FMT % collection.upper()
    else:
        return PROP_FMT % collection


class CTPPatientIdMap(dict):
    """
    CTPPatientIdMap is a dictionary augmented with a :meth:`map_subjects`
    input method to build the map and a :meth:`write` output method to
    print the CTP map properties.
    """

    SOURCE_PAT = re.compile("""
        ([a-zA-Z]+)     # The study name
        _?              # An optional underscore delimiter
        (\d+)$          # The patient number
    """, re.VERBOSE)
    """
    The input Patient ID pattern is the study name followed by a number,
    e.g. ``Breast10``.
    """

    CTP_FMT = '%s-%04d'
    """
    The CTP Patient ID format with arguments (CTP collection name, input
    Patient ID number).
    """

    MAP_FMT = 'ptid/%s=%s'
    """
    The ID lookup entry format with arguments (input Paitent ID, CTP
    patient id).
    """

    MSG_FMT = 'Mapped the QIN patient id %s to the CTP subject id %s.'
    """
    The log message format with arguments (input Paitent ID, CTP patient id).
    """

    def add_subjects(self, collection, *patient_ids):
        """
        Adds the input => CTP Patient ID association for the given input
        DICOM patient ids.

        :param collection: the image collection name 
        :param patient_ids: the DICOM Patient IDs to map
        :raise StagingError: if an input patient id format is not the study
            followed by the patient number
        """
        ctp_coll = ctp_collection_for_name(collection)
        for in_pt_id in patient_ids:
            match = CTPPatientIdMap.SOURCE_PAT.match(in_pt_id)
            if not match:
                raise StagingError("Unsupported input QIN patient id format:"
                                   " %s" % in_pt_id)
            pt_nbr = int(match.group(2))
            ctp_id = CTPPatientIdMap.CTP_FMT % (ctp_coll, pt_nbr)
            self[in_pt_id] = ctp_id
            logger(__name__).debug(
                CTPPatientIdMap.MSG_FMT % (in_pt_id, ctp_id))

    def write(self, dest=sys.stdout):
        """
        Writes this id map in the standard CTP format.
        
        :param dest: the IO stream on which to write this map (default stdout)
        """
        for qin_id in sorted(self.iterkeys()):
            print >> dest, CTPPatientIdMap.MAP_FMT % (qin_id, self[qin_id])



"""
OHSU - ROI utility functions.

TODO - move this to ohsu-qipipe.
"""

import os
import re
import glob
import qiutil


class ROIError(Exception):
    pass


class LesionROI(object):
    """
    Aggregate with attributes :attr:`lesion`, :attr:`slice` and :attr:`location`.
    """
    def __init__(self, lesion, slice_sequence_number, location):
        """
        :param lesion: the :attr:`lesion` value
        :param slice_number: the :attr:`slice` value
        :param location: the :attr:`location` value
        """
        self.lesion = lesion
        """The lesion number."""
        
        self.slice = slice_sequence_number
        """The one-based slice sequence number."""
        
        self.location = location
        """The absolute BOLERO ROI .bqf file path."""

    def __repr__(self):
        return (self.__class__.__name__ +
                str(dict(lesion=self.lesion, slice=self.slice,
                         location=self.location)))


def iter_roi(glob, regex, input_dir):
    """
    Iterates over the the OHSU ROI ``.bqf`` mask files in the given
    input directory. This method is a :class:`LesionROI` generator,
    e.g.::

        >>> # Find .bqf files anywhere under /path/to/session/processing.
        >>> next(iter_roi('processing/*', '.*/\.bqf', '/path/to/session'))
        {lesion: 1, slice: 12, path: '/path/to/session/processing/rois/roi.bqf'}

    :param glob_pat: the glob match pattern
    :;param regex: the file name match regular expression
    :param input_dir: the source session directory to search
    :yield: the :class:`LesionROI` objects
    """
    finder = qiutil.file.Finder(glob, regex)
    for match in finder.match(input_dir):
        # If there is no lesion qualifier, then there is only one lesion.
        try:
            lesion_s = match.group('lesion')
        except IndexError:
            lesion_s = None
        lesion = int(lesion_s) if lesion_s else 1
        # If there is no slice index, then complain.
        slice_seq_nbr_s = match.group('slice_sequence_number')
        if not slice_seq_nbr_s:
            raise ROIError("The ROI slice could not be determined" +
                           " from the file path: %s" % path)
        slice_seq_nbr = int(slice_seq_nbr_s)
        # Prepend the base directory to the matching file path.
        path = os.path.join(input_dir, match.group(0))
        
        yield LesionROI(lesion, slice_seq_nbr, os.path.abspath(path))



"""The proprietary OHSU mask conversion workflow."""

# TODO - move this file to ohsu-qin.pipeline.

import os
import re
import tempfile
import logging
from nipype.pipeline import engine as pe
from nipype.interfaces.utility import (IdentityInterface, Function)
import qiutil
from qiutil.logging import logger
from ..interfaces import (ConvertBoleroMask, XNATUpload)
from .workflow_base import WorkflowBase
from .pipeline_error import PipelineError

ROI_RESOURCE = 'roi'
"""The XNAT ROI resource name."""

ROI_FNAME_PAT = "lesion%d"
"""The ROI file name pattern."""


def run(project, subject, session, scan, time_series, *inputs, **opts):
    """
    Runs the ROI workflow on the given session ROI mask files.

    :param project: the project name
    :param subject: the subject name
    :param session: the session name
    :param scan: the scan number
    :param time_series: the 4D scan time series
    :param inputs: the :meth:`ROIWorkflow.run`
        (lesion number, slice sequence number, in_file) inputs
    :param opts: the :class:`ROIWorkflow` initializer options
    :return: the :meth:`ROIWorkflow.run` result
    """
    # Make the workflow.
    roi_wf = ROIWorkflow(project=project, **opts)
    # Run the workflow.
    return roi_wf.run(subject, session, scan, time_series, *inputs)


class ROIWorkflow(WorkflowBase):
    """
    The ROIWorkflow class builds and executes the ROI workflow which
    converts the BOLERO mask ``.bqf`` files to NIfTI.

    The ROI workflow input consists of the *input_spec* and *iter_slice*
    nodes. The *input_spec* contains the following input fields:

    - *subject*: the subject name

    - *session*: the session name

    - *scan*: the scan number

    - *time_series*: the 4D time series file path

    - *lesion*: the lesion number
    
    The *iter_slice* contains the following input fields:

    - *slice_sequence_number*: the one-based slice sequence number

    - *in_file*: the ROI mask``.bqf`` file to convert

    The output is the 3D mask NIfTI file location. The file name
    is *lesion*\ ``.nii.gz``.
    """

    def __init__(self, **kwargs):
        """
        If the optional configuration file is specified, then the workflow
        settings in that file override the default settings.

        :param kwargs: the :class:`qipipe.pipeline.workflow_base.WorkflowBase`
            initializer keyword arguments
        """
        super(ROIWorkflow, self).__init__(logger=logger(__name__), **kwargs)
        # The child workflow.
        self.workflow = self._create_workflow(**kwargs)
        """The ROI workflow."""

    def run(self, subject, session, scan, time_series, *inputs):
        """
        Runs the ROI workflow on the given session scan images.

        :param subject: the subject name
        :param session: the session name
        :param scan: the scan number
        :param time_series: the 4D scan time series file path
        :param inputs: the input
            (lesion number, slice sequence number, in_file)
            tuples to convert
        :return: the XNAT converted ROI resource name, or None if
            there were no inputs
        """
        if not inputs:
            self.logger.debug("Skipping the %s workflow on %s %s scan %d,"
                               "since there are no inputs to convert." %
                               (self.workflow.name, subject, session, scan))
            return None
        # Set the inputs.
        self._set_inputs(subject, session, scan, time_series, *inputs)
        # Execute the workflow.
        self.logger.debug("Executing the %s workflow on %s %s scan %d..." %
                         (self.workflow.name, subject, session, scan))
        self._run_workflow(self.workflow)
        self.logger.debug("Executed the %s workflow on %s %s scan %d." %
                         (self.workflow.name, subject, session, scan))

        # Return the resource name.
        return ROI_RESOURCE

    def _set_inputs(self, subject, session, scan, time_series, *inputs):
        """
        Sets the workflow inputs.

        :param subject: the subject name
        :param session: the session name
        :param scan: the scan number
        :param time_series: the 4D scan time series
        :param inputs: the input (lesion, slice, in_file) tuples to
            convert
        """
        # Set the execution workflow inputs.
        input_spec = self.workflow.get_node('input_spec')
        input_spec.inputs.subject = subject
        input_spec.inputs.session = session
        input_spec.inputs.scan = scan
        input_spec.inputs.time_series = time_series

        # Unpack and roll up the ROI inputs into separate iterable
        # lists.
        lesions = [roi.lesion for roi in inputs]
        slice_seq_nbrs = [roi.slice for roi in inputs]
        in_files = [roi.location for roi in inputs]
        iter_dict = dict(lesion=lesions, slice_sequence_number=slice_seq_nbrs,
                         in_file=in_files)
        iterables = iter_dict.items()
        iter_slice = self.workflow.get_node('iter_slice')
        iter_slice.iterables = iterables
        # Iterate over the ROI input fields in lock-step.
        iter_slice.synchronize = True

    def _create_workflow(self, **opts):
        """
        Makes the ROI execution workflow.

        The execution workflow input is the *input_spec* node consisting of
        the following input fields:

        - *subject*: the subject name

        - *session*: the session name

        - *scan*: the scan number

        - *time_series*: the 4D scan time series

        In addition, the workflow runner has the responsibility of setting the
        ``iter_slice`` synchronized (lesion, slice_sequence_number, in_file)
        iterables.

        :param opts: the workflow creation options:
        :return: the execution workflow
        """
        self.logger.debug("Creating the ROI execution workflow...")

        # The execution workflow.
        exec_wf = pe.Workflow(name='roi_exec', base_dir=self.base_dir)

        # The ROI workflow input.
        input_fields = ['subject', 'session', 'scan', 'time_series', 'resource']
        input_spec = pe.Node(IdentityInterface(fields=input_fields),
                             name='input_spec')
        input_spec.inputs.resource = ROI_RESOURCE

        # The input ROI tuples are iterable.
        iter_slice_fields = ['lesion', 'slice_sequence_number', 'in_file']
        iter_slice = pe.Node(IdentityInterface(fields=iter_slice_fields),
                           name='iter_slice')

        # The merged 3D output file base name.
        basename_xfc = Function(input_names=['lesion'],
                                output_names=['basename'],
                                function=base_name)
        basename = pe.Node(basename_xfc, name='basename')
        exec_wf.connect(iter_slice, 'lesion', basename, 'lesion')

        # Convert the input file.
        convert = pe.Node(ConvertBoleroMask(), name='convert')
        exec_wf.connect(iter_slice, 'in_file', convert, 'in_file')
        exec_wf.connect(iter_slice, 'slice_sequence_number',
                        convert, 'slice_sequence_number')
        exec_wf.connect(input_spec, 'time_series', convert, 'time_series')
        exec_wf.connect(basename, 'basename', convert, 'out_base')

        # Merge the slices.
        merge = pe.JoinNode(MergeNifti(), joinsource='iter_slice',
                            joinfield='in_files', name='merge_slices')
        exec_wf.connect(convert, 'out_file', merge, 'in_files')
        exec_wf.connect(basename, 'basename', merge, 'out_format')

        # Upload the ROI result into the XNAT ROI resource.
        upload_roi_xfc = XNATUpload(project=self.project,
                                    resource=ROI_RESOURCE, modality='MR')
        upload_roi = pe.Node(upload_roi_xfc, name='upload_roi')
        exec_wf.connect(input_spec, 'subject', upload_roi, 'subject')
        exec_wf.connect(input_spec, 'session', upload_roi, 'session')
        exec_wf.connect(input_spec, 'scan', upload_roi, 'scan')
        exec_wf.connect(merge, 'out_file', upload_roi, 'in_files')

        self.logger.debug("Created the %s workflow." % exec_wf.name)
        # If debug is set, then diagram the workflow graph.
        if self.logger.level <= logging.DEBUG:
            self.depict_workflow(exec_wf)

        return exec_wf


### Utility functions called by the workflow nodes. ###

def base_name(lesion):
    """
    :param lesion: the lesion number
    :return: the base name to use
    """
    return ROI_FNAME_PAT % lesion



from qiprofile_rest_client.helpers import database
from qiprofile_rest_client.model.subject import Subject
from qiprofile_rest_client.model.imaging import Session
from . import (clinical, imaging)


def update(project, collection, subject, session, spreadsheet):
    """
    Updates the qiprofile database from the clinical spreadsheet and
    XNAT database for the given session.

    :param project: the XNAT project name
    :param collection: the image collection name
    :param subject: the subject number
    :param session: the XNAT session number
    :param spreadsheet: the spreadsheet input file location
    :param modeling_technique: the modeling technique
    """
    # Get or create the subject database subject.
    key = dict(project=project, collection=collection, number=subject)
    sbj = database.get_or_create(Subject, key)
    # Update the clinical information from the XLS input.
    clinical.update(sbj, spreadsheet)
    # Update the imaging information from XNAT.
    imaging.update(sbj, session)



"""
OHSU - This module wraps the proprietary OHSU AIRC ``fastfit`` software.
``fastfit`` optimizes the input pharmacokinetic model.

:Note: this interface is copied from the AIRC cluster file
    ``/usr/global/scripts/fastfit_iface.py``. It is included
    in :mod:`qipipe.interfaces` in order to gather all custom OHSU
    QIN Python source code in the ``qipipe`` module. If it were not
    included, then qipipe would not compile in a non-AIRC cluster
    environment, even if this ``FastFit`` interface is unused.
"""

# TODO - move this to the ohsu-qin.pipeline.interfaces module

# Absolute import (the default in a future Python release) resolves
# the fastfit import as the installed fastfit Python module rather
# than this interface of the same name.
from __future__ import absolute_import
import os
from os import path
from glob import glob
import traits.api as traits
from nipype.interfaces.base import (DynamicTraitedSpec,
                                    MpiCommandLine,
                                    MpiCommandLineInputSpec,
                                    isdefined)
from nipype.interfaces.traits_extension import Undefined
from fastfit.fastfit_cli import get_available_models
from .interface_error import InterfaceError


class FastfitError(Exception):
    pass


class FastfitInputSpec(MpiCommandLineInputSpec):
    model_name = traits.String(desc='The name of the model to optimize',
                               mandatory=True, position=-2, argstr='%s')

    target_data = traits.File(desc='Target data',
                              mandatory=True, position=-1, argstr='%s')

    mask = traits.File(desc='Mask file', argstr='-m %s')

    weights = traits.File(desc='Weights file', argstr='-w %s')

    params = traits.Dict(desc='Parameters for the model')

    params_csv = traits.File(desc='Parameters CSV', argstr='--param-csv %s')

    fix_params = traits.Dict(desc="Optimization parameters to fix, and "
                             "the values to fix them to", argstr='%s')

    optional_outs = traits.List(desc='Optional outputs to produce',
                                argstr='%s')


class Fastfit(MpiCommandLine):
    """
    Interface to the ``fastfit`` software package.

    Input parameters:
    
    * *min_outs*: The minimum outputs expected. Required if the *model_name*
      input is set dynamically.
    """

    _cmd = 'fastfit'
    input_spec = FastfitInputSpec
    output_spec = DynamicTraitedSpec

    def __init__(self, min_outs=None, **inputs):
        super(Fastfit, self).__init__(**inputs)
        self._min_outs = min_outs

    def _format_arg(self, name, spec, value):
        if name == 'optional_outs':
            return ' '.join("-o '%s'" % opt_out for opt_out in value)
        elif name == 'params':
            return ' '.join(["-s %s:%s" % item
                             for item in value.iteritems()])
        elif name == 'fix_params':
            return ' '.join(["--fix-param %s:%s" % item
                             for item in value.iteritems()])
        else:
            return spec.argstr % value

    def _outputs(self):
        #Set up dynamic outputs for resulting parameter maps
        outputs = super(Fastfit, self)._outputs()
        undefined_traits = {}

        #Get a list of fixed params
        fixed_params = []
        if isdefined(self.inputs.fix_params):
            for key in self.inputs.fix_params:
                fixed_params.append(key)

        #If the model name is static, we can find the outputs
        if isdefined(self.inputs.model_name):
            model_mods, builtin_models, user_models = get_available_models()
            model = model_mods[self.inputs.model_name]
            self._opt_params = model.optimization_params
            if (not self._min_outs is None and
                any(not out in self._opt_params
                    for out in self._min_outs)):
                raise FastfitError("The model %s does not provide the "
                                 "minimum outputs" % model_name)
            for param_name in self._opt_params:
                if not param_name in fixed_params:
                    outputs.add_trait(param_name, traits.File(exists=True))
                    undefined_traits[param_name] = Undefined

        #Otherwise we need min_outs
        elif not self._min_outs is None:
            for param_name in self._min_outs:
                outputs.add_trait(param_name, traits.File(exists=True))
                undefined_traits[param_name] = Undefined
        else:
            raise FastfitError("Either the 'model_name' input must "
                               "be static or the 'min_outs' argument "
                               "to the constructor must be given.")

        #Set up dynamic outputs for any requested optional outputs
        if isdefined(self.inputs.optional_outs):
            for opt_out in self.inputs.optional_outs:
                outputs.add_trait(opt_out, traits.File(exists=True))
                undefined_traits[opt_out] = Undefined

        outputs.trait_set(trait_change_notify=False, **undefined_traits)
        for dynamic_out in undefined_traits.keys():
            _ = getattr(outputs, dynamic_out) #Not sure why, but this is needed
        return outputs

    def _list_outputs(self):
        outputs = self._outputs().get()

        cwd = os.getcwd()
        for param_name in outputs.keys():
            outputs[param_name] = path.join(cwd, '%s_map.nii.gz' % param_name)

        return outputs



"""
Image processing preparation.

The staging package defines the functions used to prepare the study image files
for import into XNAT, submission to the TCIA QIN collections and pipeline
processing.
"""

# OHSU - The ohsu module creates the OHSU QIN collections.
# TODO - this should be a config item.
from . import ohsu



import os
import dicom
from matplotlib import (image, cm)
from nipype.interfaces.base import (traits, BaseInterfaceInputSpec,
                                    TraitedSpec, BaseInterface)


class PreviewInputSpec(BaseInterfaceInputSpec):
    in_file = traits.File(exists=True, mandatory=True,
                          desc='The input 3D NIfTI file')

    dest = traits.Directory(desc='The destination directory path'
                                 ' (default current directory)')
    
    crop = traits.List(traits.List(traits.Int),
                       desc='The x and y [min, max] bounds')

    out_fname = traits.File(desc='The destination file name'
                                 ' (default is the input file name)')


class PreviewOutputSpec(TraitedSpec):
    out_file = traits.File(exists=True, desc='The preview JPEG file')


class Preview(BaseInterface):
    """Preview creates a JPEG image from an input DICOM image."""

    input_spec = PreviewInputSpec

    output_spec = PreviewOutputSpec

    def _run_interface(self, runtime):
        self._out_file = self._convert(
            self.inputs.in_file, dest=self.inputs.dest,
            out_fname=self.inputs.out_fname, crop=self.inputs.crop
        )
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_file'] = self._out_file

        return outputs

    def _convert(self, in_file, dest=None, out_fname=None, crop=None):
        """
        Copies the given file.
    
        :param in_file: the path of the file or directory to copy
        :param dest: the destination directory path
            (default is the current directory)
        :param out_fname: the destination file name
            (default is the input file name)
        :param crop: the x and y [min, max] bounds
        :return: the preview JPEG file path
        """
        if dest:
            dest = os.path.abspath(dest)
            if not os.path.exists(dest):
                os.makedirs(dest)
        else:
            dest = os.getcwd()

        # The default output file name is the input file name
        # with a .jpg extension.
        if not out_fname:
            _, in_fname = os.path.split(in_file)
            base, _ = os.path.splitext(in_fname)
            out_fname = base + '.jpg'
        out_file = os.path.join(dest, out_fname)
        
        dcm = dicom.read_file(in_file)
        data = dcm.pixel_array
        # Crop the data, if necessary.
        if crop:
            crop_x, crop_y = crop
            crop_xmin, crop_xmax = crop_x
            crop_ymin, crop_ymax = crop_y
            data = data[crop_xmin:crop_xmax, crop_ymin:crop_ymax]
        image.imsave(out_file, data, cmap=cm.jet)
        
        return out_file



from nipype.interfaces.base import (traits, DynamicTraitedSpec, isdefined)
from nipype.interfaces.io import (IOBase, add_traits)
from .interface_error import InterfaceError


class Unpack(IOBase):
    """
    The Unpack Interface converts a list input field to one output field per list item.

    Example:

    >>> from qipipe.interfaces.unpack import Unpack
    >>> unpack = Unpack(input_name='list', output_names=['a', 'b'], list=[1, 2])
    >>> result = unpack.run()
    >>> result.outputs.a
    1
    >>> result.outputs.b
    2
    """

    input_spec = DynamicTraitedSpec

    output_spec = DynamicTraitedSpec

    def __init__(self, input_name, output_names, mandatory_inputs=True, **inputs):
        """
        :param input_name: the input list field name
        :param output_names: the output field names
        :param mandatory_inputs: a flag indicating whether every input field is
            required
        :param inputs: the input field name => value bindings
        """
        super(Unpack, self).__init__(**inputs)
        if not input_name:
            raise InterfaceError('Unpack input field name is missing')
        if not output_names:
            raise InterfaceError('Unpack output field names must be a non-empty list')
        self.input_names = [input_name]
        self.output_names = output_names
        self._mandatory_inputs = mandatory_inputs
        add_traits(self.inputs, self.input_names, traits.List)
        # Adding any traits wipes out all input values set in superclass
        # initialization, even it the trait is not in the add_traits argument.
        # The work-around is to reset the values after adding the traits.
        self.inputs.set(**inputs)

    def _add_output_traits(self, base):
        return add_traits(base, self.output_names)

    def _run_interface(self, runtime):
        # Manual mandatory inputs check.
        if self._mandatory_inputs:
            for key in self.input_names:
                value = getattr(self.inputs, key)
                if not isdefined(value):
                    msg = ("%s requires a value for input '%s' because it was"
                        " listed in 'input_names'. You can turn off mandatory"
                        " inputs checking  by passing mandatory_inputs = False"
                        " to the constructor." % (self.__class__.__name__, key))
                    raise InterfaceError(msg)

        # The input list.
        in_list = getattr(self.inputs, self.input_names[0])
        # The output name => value dictionary result.
        self._result = {self.output_names[i]: value for i, value in enumerate(in_list)}

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        for key, value in self._result.iteritems():
            outputs[key] = value
        return outputs



"""
This module updates the qiprofile database Subject chemotherapy protocol
information from a Chemotherapy Excel worksheet.
"""

from qiprofile_rest_client.model.clinical import (Dosage, Drug)
from .dosage import (DosageWorksheet, DosageUpdate)
from . import parse

COL_ATTRS = {'Cumulative Amount (mg/m2 BSA)': 'amount'}
"""
The following non-standard column-attribute associations:
* The Cumulative Amount column is the amount attribute.
"""

SHEET = 'Chemotherapy'
"""The input XLS sheet name."""


def read(workbook, **condition):
    """
    This is a convenience method that wraps :class:`ChemotherapyWorksheet`
    :meth:`qipipe.qiprofile.xls.Worksheet.read`.

    :param workbook: the read-only ``openpyxl`` workbook object 
    :param condition: the :meth:`qipipe.qiprofile.xls.Worksheet.read`
        filter condition
    :return: the :meth:`qipipe.qiprofile.xls.Worksheet.read` rows
    """
    reader = DosageWorksheet(workbook, SHEET, Drug,
                             column_attributes=COL_ATTRS)

    return reader.read(**condition)


def update(subject, rows):
    """
    Updates the given subject data object from the dosage XLS rows.

    :param subject: the ``Subject`` Mongo Engine database object
        to update
    :param rows: the input chemotherapy :meth:`read` rows list 
    
    """
    updater = DosageUpdate(subject, Drug)
    updater.update(rows)



# Absolute import (the default in a future Python release) resolves
# the collections import as the standard Python collections module
# rather than the staging collections module.
from __future__ import absolute_import
import os
import re
import glob
from collections import defaultdict
from qiutil.logging import logger
import qixnat
import qidicom.hierarchy
from . import collections
from ..helpers.constants import (SUBJECT_FMT, SESSION_FMT)
from .roi import iter_roi
from .staging_error import StagingError


class ScanInput(object):
    def __init__(self, collection, subject, session, scan, iterators):
        self.collection = collection
        """The image collection name."""

        self.subject = subject
        """The scan subject name."""
        
        self.session = session
        """The scan session name."""
        
        self.scan = scan
        """The scan number."""
        
        self.iterators = iterators
        """The :class:`ScanIterators` object."""

    def __repr__(self):
        return (self.__class__.__name__ +
                str(dict(collection=self.collection, subject=self.subject,
                         scan=self.scan, iterators=self.iterators)))


class ScanIterators(object):
    """
    Aggregate with attributes :attr:`location`, :meth:`dicom` and :meth:`roi`.
    """

    def __init__(self, pattern, dicom_gen, roi_gen=None):
        """
        :param pattern: the scan file pattern
        :param dicom_gen: the :attr:`dicom` generator
        :param roi_gen: the :attr:`roi` generator
        """
        self.pattern = pattern
        self._dicom_gen = dicom_gen
        self._dicom = None
        self._roi_gen = roi_gen
        self._roi = None
    
    @property
    def dicom(self):
        """The {volume: [DICOM files]} dictionary."""
        if self._dicom == None:
            # The {volume: iterator} dictionary.
            dicom_iters = next(self._dicom_gen)
            # The {volume: list} dictionary.
            dicom_lists = {k: list(v) for k, v in dicom_iters.iteritems()}
            # Remove the empty file lists.
            self._dicom = {k: v for k, v in dicom_lists.iteritems() if v}
        return self._dicom
    
    @property
    def roi(self):
        """The :meth:`qipipe.staging.roi.iter_roi` iterator."""
        if self._roi == None:
            self._roi = list(self._roi_gen) if self._roi_gen else []
        return self._roi

    def __repr__(self):
        return (self.__class__.__name__ +
                str(dict(dicom=self.dicom, roi=self.roi)))


def iter_stage(project, collection, *inputs, **opts):
    """
    Iterates over the the new sessions in the given input directories.
    This method is a staging generator which yields a tuple consisting
    of the subject, session, scan number and :class:`ScanIterators`
    object, e.g.::
    
        >> scan_input = next(iter_stage('QIN', 'Sarcoma', '/path/to/subject'))
        >> print scan_input.subject
        Sarcoma001
        >> print scan_input.session
        Session01
        >> print scan_input.scan
        1
        >> print scan_input.iterators.dicom
        {1: ['/path/to/t1/image1.dcm', ...], ...}
        >> print scan_input.iterators.roi
        [(1, 19, '/path/to/roi19.bqf'), ...]

    The input directories conform to the
    :attr:`qipipe.staging.collection.Collection.patterns`
    :attr:`qipipe.staging.collection.Patterns.subject`
    regular expression.

    Each iteration *(subject, session, scan, scan_iters)* tuple is
    formed as follows:

    - The *subject* is the XNAT subject name formatted by
      :data:`SUBJECT_FMT`.

    - The *session* is the XNAT experiment name formatted by
      :data:`SESSION_FMT`.

    - The *scan* is the XNAT scan number.

    - The *scan_iters* is the :class:`ScanIterators` object.

    :param project: the XNAT project name
    :param collection: the
        :attr:`qipipe.staging.collection.Collection.name`
    :param inputs: the source subject directories to stage
    :param opts: the following keyword option:
    :keyword scan: the scan number to stage
        (default stage all detected scans)
    :keyword skip_existing: flag indicating whether to ignore each
        existing session, or scan if the *scan* option is set
        (default True)
    :yield: the :class:`ScanInput` object
    """
    # Validate that there is a collection
    if not collection:
        raise StagingError('Staging is missing the image collection name')
    
    # Group the new DICOM files into a
    # {subject: {session: {scan: scan iterators}} dictionary.
    stg_dict = _collect_visits(project, collection, *inputs, **opts)

    # Generate the (subject, session, :class:ScanIterators) tuples.
    _logger = logger(__name__)
    for sbj, sess_dict in stg_dict.iteritems():
        for sess, scan_dict in sess_dict.iteritems():
            for scan, scan_iters in scan_dict.iteritems():
                # The scan must have at least one DICOM file.
                if scan_iters.dicom:
                    _logger.debug("Staging %s %s scan %d..." % (sbj, sess, scan))
                    yield ScanInput(collection, sbj, sess, scan, scan_iters)
                    _logger.debug("Staged %s %s scan %d." % (sbj, sess, scan))
                else:
                    _logger.debug("Skipping %s %s scan %d since no DICOM files"
                                  " were found for this scan matching %s." %
                                  (sbj, sess, scan, scan_iters.pattern))


def _collect_visits(project, collection, *inputs, **opts):
    """
    Collects the sessions in the given input directories. The session
    DICOM files are grouped by volume.

    :param project: the XNAT project name
    :param collection: the TCIA image collection name
    :param inputs: the source DICOM subject directories
    :param opts: the :meth:`iter_stage` options
    :return: the {subject: {session: {scan: :class:`ScanIterators`}}}
        dictionary
    """
    # The visit (subject, session, scan dictionary) tuple generator.
    visits = VisitIterator(project, collection, *inputs, **opts)

    # The dictionary to build.
    visit_dict = defaultdict(dict)
    # Add each tuple as a dictionary entry.
    for sbj, sess, scan_dict in visits:
        visit_dict[sbj][sess] = scan_dict
    
    return visit_dict


class VisitIterator(object):
    """Scan DICOM generator class ."""

    def __init__(self, project, collection, *subject_dirs, **opts):
        """
        :param project: the XNAT project name
        :param collection: the TCIA image collection name
        :param subject_dirs: the subject directories over which
            to iterate
        :param opts: the :meth:`iter_stage` options
        """
        self.project = project
        """The :meth:`iter_stage` project name parameter."""
        
        self.collection = collections.with_name(collection)
        """The :meth:`iter_stage` collection name parameter."""

        self.subject_dirs = subject_dirs
        """The input directories."""

        self.scan = opts.get('scan')
        """The :meth:`iter_stage` scan number option."""

        self.skip_existing = opts.get('skip_existing', True)
        """The :meth:`iter_stage` *skip_existing* flag option."""
        
        self.logger = logger(__name__)

    def __iter__(self):
        """
        Returns the next (subject, session, scan_dict) tuple for the
        scans in the subject directories, where:
        * *subject* is the subject name
        * *session* is the session name
        * *scan_dict* is the {scan number: :class:`ScanIterators`}
            dictionary
        
        :return: the next (subject, session, scan_dict) tuple
        """
        # The visit subdirectory matcher.
        vpat = self.collection.patterns.session
        # The {scan number: {dicom, roi}} file search patterns.
        all_scan_pats = self.collection.patterns.scan
        # The selected file search patterns.
        if self.scan:
            # Filter on only the specified scan.
            if self.scan not in all_scan_pats:
                raise StagingError("The %s scan %d is not supported" %
                                   (self.collection.name, self.scan))
            scan_pats = {self.scan: all_scan_pats[self.scan]}
        else:
            # Detect all scans.
            scan_pats = all_scan_pats

        filter_sess = self.skip_existing and not self.scan
        filter_scan = self.skip_existing and self.scan
        
        # Iterate over the visits.
        with qixnat.connect():
            for sbj_dir in self.subject_dirs:
                sbj_dir = os.path.abspath(sbj_dir)
                self.logger.debug("Discovering sessions in %s..." % sbj_dir)
                # Make the XNAT subject name.
                _, sbj_subdir = os.path.split(sbj_dir)
                sbj_nbr = self._match_subject_number(sbj_subdir)
                sbj = SUBJECT_FMT % (self.collection.name, sbj_nbr)
                # The subject subdirectory names which match the visit
                # pattern, sorted in visit order.
                sess_matches = sorted((d for d in os.listdir(sbj_dir)
                                       if vpat.match(d)))
                # Generate the new (subject, session, {scan: DICOM files})
                # tuples for each visit.
                for sess_subdir in sess_matches:
                    # The visit directory path.
                    sess_dir = os.path.join(sbj_dir, sess_subdir)
                    # Silently skip non-directories.
                    if not os.path.isdir(sess_dir):
                        continue
                    # The visit (session) number.
                    sess_nbr = self._match_session_number(sess_subdir)
                    # The XNAT session name.
                    sess = SESSION_FMT % sess_nbr
                    # Apply the selection filter, e.g. an XNAT existence
                    # check. If the session passes the filter, then the
                    # files qualify for iteration.
                    if not filter_sess or self._is_new_session(sbj, sess):
                        # The DICOM and ROI iterators for each scan number.
                        scan_dict = {scan: self._scan_iterators(pats, sess_dir)
                                     for scan, pats in scan_pats.iteritems()
                                     if not filter_scan or self._is_new_scan(sbj, sess, scan)}
                        if scan_dict:
                            scans = scan_dict.keys()
                            self.logger.debug("Discovered %s %s scans %s in %s" %
                                              (sbj, sess, scans, sess_dir))
                            yield sbj, sess, scan_dict
                        else:
                            self.logger.debug("No %s %s scans were discovered"
                                              " in %s" %  (sbj, sess, sess_dir))

    def _scan_iterators(self, patterns, input_dir):
        # The DICOM glob pattern.
        dcm_pat = os.path.join(input_dir, patterns.dicom)
        # The DICOM file generator.
        dcm_gen = _scan_dicom_generator(dcm_pat,
                                        self.collection.patterns.volume)
        # The ROI file match patterns.
        roi_pats = patterns.roi
        # Make the ROI generator, if necessary.
        if roi_pats:
            roi_gen = iter_roi(roi_pats.glob, roi_pats.regex, input_dir)
        else:
            roi_gen = None 
        
        return ScanIterators(pattern=dcm_pat, dicom_gen=dcm_gen, roi_gen=roi_gen)

    def _match_subject_number(self, path):
        """
        :param path: the directory path
        :return: the subject number
        :raise StagingError: if the path does not match the collection subject
            pattern
        """
        match = self.collection.patterns.subject.match(path)
        if not match:
            raise StagingError(
                "The directory path %s does not match the subject pattern %s" %
                (path, self.collection.patterns.subject.pattern))

        return int(match.group(1))

    def _match_session_number(self, path):
        """
        :param path: the directory path
        :return: the session number
        :raise StagingError: if the path does not match the collection session
            pattern
        """
        match = self.collection.patterns.session.match(path)
        if not match:
            raise StagingError(
                "The directory path %s does not match the session pattern %s" %
                (path, self.collection.patterns.session.pattern))
        return int(match.group(1))
    
    def _is_new_session(self, subject, session):
        with qixnat.connect() as xnat:
            sess = xnat.find_one(self.project, subject, session)
        if sess:
            logger(__name__).debug("Skipping %s %s since it has already been"
                                   " loaded to XNAT." % (subject, session))
        return not sess

    def _is_new_scan(self, subject, session, scan):
        with qixnat.connect() as xnat:
            scan_obj = xnat.find_one(self.project, subject, session, scan=scan)
        if scan_obj:
            logger(__name__).debug("Skipping %s %s scan %d since it has"
                                   " already been loaded to XNAT." %
                                   (subject, session, scan))
        return not scan_obj


def _scan_dicom_generator(pattern, tag):
    """
    :param pattern: the DICOM file glob pattern
    :param tag: the DICOM volume tag
    :yield: the {volume: [DICOM files]} dictionary 
    """
    # The visit directory DICOM file iterator.
    dicom_files = glob.iglob(pattern)
    
    # Group the DICOM files by volume.
    yield qidicom.hierarchy.group_by(tag, *dicom_files)

        


"""
This module updates the qiprofile database Subject demographics
information from the demographics Excel workbook file.
"""

import re
from qiprofile_rest_client.model.subject import Subject
from .xls import Worksheet
from . import parse

SHEET = 'Demographics'
"""The input XLS sheet name."""

COL_ATTRS = {'Race': 'races'}
"""
The following non-standard column-attribute associations:
* The Race column is the races attribute.
"""


class DemographicsError(Exception):
    pass


def read(workbook, **condition):
    """
    Reads the demographics XLS row which matches the given subject.

    :param condition: the row selection filter
    :return: the Demographics sheet
        :meth:`qipipe.qiprofile.xls.Worksheet.read`
        row bundle which matches the given subject, or None if no
        match was found
    """
    # Wrap the worksheet to read Subject attributes.
    reader = Worksheet(workbook, SHEET, Subject, column_attributes=COL_ATTRS)
    
    # The filtered worksheet rows.
    return reader.read(**condition)


def update(subject, rows):
    """
    Updates the given subject data object from the given Demographics
    sheet rows.
    
    There can be no more than one Demographics update input row for
    the given subject. The *rows* parameter is an iterable in order to
    conform to other sheet facade modules.

    :param subject: the ``Subject`` Mongo Engine database object
        to update
    :param rows: the input Demographics :meth:`read` rows
    :raise DemographicsError: if there is more than one input row
    """
    row_list = list(rows)
    # An empty row is valid and is a no-op.
    # More than one row is an error.
    if not row_list:
        return
    elif len(row_list) > 1:
        raise DemographicsError("Subject number %d has more than one row" %
                                 subject.number)
    # Use the single row.
    row = row_list[0]
    for attr, val in row.iteritems():
        if hasattr(subject, attr):
            setattr(subject, attr, val)



import os
import re
import tempfile
import logging
from nipype.pipeline import engine as pe
from nipype.interfaces.utility import (IdentityInterface, Function, Merge)
from nipype.interfaces.dcmstack import MergeNifti
import qixnat
from qixnat.helpers import path_hierarchy
from qiutil.logging import logger
from . import (staging, registration, modeling)
from .pipeline_error import PipelineError
from .workflow_base import WorkflowBase
from .staging import StagingWorkflow
from .modeling import (ModelingWorkflow, MODELING_PROFILE_FILE)
from .mask import MaskWorkflow
from .roi import ROIWorkflow
from ..interfaces import (XNATDownload, XNATUpload)
from ..staging.iterator import iter_stage
from ..staging.map_ctp import map_ctp
# OHSU - multi-volume scans.
# TODO - get this from a [Staging] multi_volume_scans config parameter.,
# with default all scans.
from ..staging.ohsu import MULTI_VOLUME_SCAN_NUMBERS

SCAN_TS_RSC = 'scan_ts'
"""The XNAT scan time series resource name."""

MASK_RSC = 'mask'
"""The XNAT mask resouce name."""

ACTIONS = ['stage', 'roi', 'qiprofile', 'register', 'model']
"""The list of all workflow actions."""

VOLUME_FILE_PAT = re.compile("volume(\d{3}).nii.gz$")
"""
The volume file name pattern. The image file name is
volume<number>.nii.gz, where <number> is the zero-padded volume
number, as determined by the
:meth:`qipipeline.pipeline.staging.volume_format` function.
"""


def run(*inputs, **opts):
    """
    Creates a :class:`qipipe.pipeline.qipipeline.QIPipelineWorkflow`
    and runs it on the given inputs. The pipeline execution depends
    on the *actions* option, as follows:

    - If the workflow actions includes ``stage`` or ``roi``, then
      the input is the :meth:`QIPipelineWorkflow.run_with_dicom_input`
      DICOM subject directories input.

    - Otherwise, the input is the
      :meth:`QIPipelineWorkflow.run_with_scan_download` XNAT session
      labels input.

    :param inputs: the input directories or XNAT session labels to
        process
    :param opts: the :meth:`qipipe.staging.iterator.iter_stage`
        and :class:`QIPipelineWorkflow` initializer options,
        as well as the following keyword options:
    :keyword project: the XNAT project name
    :keyword collection: the image collection name
    :keyword actions: the workflow actions to perform
        (default :const:`ACTIONS`)
    :keyword resume: flag indicating whether to resume processing on
        existing sessions (default False)
    """
    # The actions to perform.
    actions = opts.pop(
    'actions', ACTIONS)
    if 'stage' in actions:
        # Run with staging DICOM subject directory input.
        _run_with_dicom_input(actions, *inputs, **opts)
    elif 'roi' in actions:
        # The non-staging ROI action must be performed alone.
        if len(actions) > 1:
            raise ArgumentError("The ROI pipeline can only be run with"
                                " staging or stand-alone")
        _run_with_dicom_input(actions, *inputs, **opts)
    else:
        # Run downstream actions with XNAT session input.
        _run_with_xnat_input(actions, *inputs, **opts)


def _run_with_dicom_input(actions, *inputs, **opts):
    """
    :param actions: the actions to perform
    :param inputs: the DICOM directories to process
    :param opts: the :meth:`run` options
    """
    # The required XNAT project name.
    project = opts.pop('project', None)
    if not project:
        raise ArgumentError('The staging pipeline project option is missing.')
    # The required image collection name.
    collection_opt = opts.pop('collection', None)
    if not collection_opt:
        raise ArgumentError('The staging pipeline collection option is missing.')
    # The collection name is capitalized.
    collection = collection_opt.capitalize()
    # The target directory.
    dest = opts.get('dest', None)
    # The resume option corresponds to the staging helper iter_stage
    # function skip_existing option.
    if opts.pop('resume', None):
        opts['skip_existing'] = False

    # The set of input subjects is used to build the CTP mapping file
    # after the workflow is completed, if staging is enabled.
    subjects = set()
    # Run the workflow on each session and scan.
    for scan_input in iter_stage(project, collection, *inputs, **opts):
        # OHSU - Only multi-volume scans can have post-staging downstream
        # actions.
        if scan_input.scan in MULTI_VOLUME_SCAN_NUMBERS:
            if 'stage' in actions:
                wf_actions = ['stage']
            else:
                self.logger.debug("Skipping %s %s %s scan %d, since the"
                            " actions do not include stage and only"
                            " multi-volume scans support post-stage"
                            "actions." %
                            (project, scan_input.subject, scan_input.session,
                             scan_input.scan))
                continue
        else:
            wf_actions = actions
        # Capture the subject.
        subjects.add(scan_input.subject)
        # Create a new workflow.
        wf_gen = QIPipelineWorkflow(project, wf_actions, **opts)
        # Run the workflow on the scan.
        wf_gen.run_with_dicom_input(wf_actions, scan_input, dest)

    # If staging is enabled, then make the TCIA subject map.
    if 'stage' in actions:
        map_ctp(collection, *subjects, dest=dest)


def _run_with_xnat_input(actions, *inputs, **opts):
    """
    Run the pipeline with a XNAT download. Each input is a XNAT scan
    path, e.g. ``/QIN/Breast012/Session03/scan/1``.

    :param actions: the actions to perform
    :param inputs: the XNAT scan resource paths
    :param opts: the :class:`QIPipelineWorkflow` initializer options
    """
    for path in inputs:
        hierarchy = dict(path_hierarchy(path))
        prj = hierarchy.pop('project', None)
        if not prj:
            raise ArgumentError("The XNAT path is missing a project: %s" % path)
        sbj = hierarchy.pop('subject', None)
        if not sbj:
            raise ArgumentError("The XNAT path is missing a subject: %s" % path)
        sess = hierarchy.pop('experiment', None)
        if not sess:
            raise ArgumentError("The XNAT path is missing a session: %s" % path)
        scan_s = hierarchy.pop('scan', None)
        if not scan_s:
            raise ArgumentError("The XNAT path is missing a scan: %s" % path)
        scan = int(scan_s)
        # The XNAT connection is open while the input scan is processed.
        with qixnat.connect() as xnat:
            # The XNAT scan object must exist.
            scan_obj = xnat.find_one(prj, sbj, sess, scan=scan)
            if not scan_obj:
                raise ArgumentError("The XNAT scan object does not exist: %s" % path)

            # The workflow options are augmented from the base options.
            wf_opts = dict(opts)
            # Check for an existing mask.
            if _scan_resource_exists(xnat, prj, sbj, sess, scan, MASK_RSC):
                wf_opts['mask'] = MASK_RSC

            # Every post-stage action requires a 4D scan time series.
            ts_actions = (action for action in actions if action != 'stage')
            if any(ts_actions):
                if _scan_resource_exists(xnat, prj, sbj, sess, scan,
                                         SCAN_TS_RSC):
                    wf_opts['scan_time_series'] = SCAN_TS_RSC

            # If modeling will be performed on a specified registration
            # resource, then check for an existing 4D registration time
            # series.
            reg_rsc_opt = opts.get('registration_resource')
            if 'model' in actions and reg_rsc_opt:
                reg_ts_rsc = reg_rsc_opt + '_ts'
                if _scan_resource_exists(xnat, prj, sbj, sess, scan,
                                         reg_ts_rsc):
                    wf_opts['realigned_time_series'] = reg_ts_rsc

            # Execute the workflow.
            wf_gen = QIPipelineWorkflow(prj, actions, **wf_opts)
            # If a registration resource was specified, then set the flag
            # to check for registered scans.
            check_reg = opts.get('registration_resource') != None
            wf_gen.run_with_scan_download(xnat, prj, sbj, sess, scan, actions,
                                          check_reg)


def _scan_resource_exists(xnat, project, subject, session, scan, resource):
    """
    :return: whether the given XNAT scan resource exists
    """
    rsc_obj = xnat.find_one(project, subject, session, scan=scan,
                        resource=resource)
    exists = rsc_obj and rsc_obj.files().get()
    status = 'found' if exists else 'not found'
    logger(__name__).debug("The %s %s %s scan %d resource %s was %s." %
                           (project, subject, session, scan, resource,
                            status))

    return exists


class ArgumentError(Exception):
    pass


class NotFoundError(Exception):
    pass


class QIPipelineWorkflow(WorkflowBase):
    """
    QIPipeline builds and executes the imaging workflows. The pipeline
    builds a composite workflow which stitches together the following
    constituent workflows:

    - staging: Prepare the new DICOM visits, as described in
      :class:`qipipe.pipeline.staging.StagingWorkflow`

    - mask: Create the mask from the staged images,
      as described in :class:`qipipe.pipeline.mask.MaskWorkflow`

    - registration: Mask, register and realign the staged images,
      as described in
      :class:`qipipe.pipeline.registration.RegistrationWorkflow`

    - modeling: Perform PK modeling as described in
      :class:`qipipe.pipeline.modeling.ModelingWorkflow`

    The constituent workflows are determined by the initialization
    options ``stage``, ``register`` and ``model``. The default is
    to perform each of these subworkflows.

    The workflow steps are determined by the input options as follows:

    - If staging is performed, then the DICOM files are staged for the
      subject directory inputs. Otherwise, staging is not performed.
      In that case, if registration is enabled as described below, then
      the previously staged volume scan stack images are downloaded.

    - If registration is performed and the ``registration`` resource
      option is set, then the previously realigned images with the
      given resource name are downloaded. The remaining volumes are
      registered.

    - If registration or modeling is performed and the XNAT ``mask``
      resource is found, then that resource file is downloaded.
      Otherwise, the mask is created from the staged images.

    The workflow input node is *input_spec* with the following
    fields:

    - *subject*: the subject name

    - *session*: the session name

    - *scan*: the scan number

    In addition, if the staging or registration workflow is enabled
    then the *iter_volume* node iterables input includes the
    following fields:

    - *volume*: the scan number

    - *dest*: the target staging directory, if the staging
      workflow is enabled

    The constituent workflows are combined as follows:

    - The staging workflow input is the workflow input.

    - The mask workflow input is the newly created or previously staged
      scan NIfTI image files.

    - The modeling workflow input is the combination of the previously
      uploaded and newly realigned image files.

    The pipeline workflow is available as the
    :attr:`qipipe.pipeline.qipipeline.QIPipelineWorkflow.workflow`
    instance variable.
    """

    def __init__(self, project, actions, **opts):
        """
        :param project: the XNAT project name
        :param actions: the actions to perform
        :param opts: the :class:`qipipe.staging.WorkflowBase`
            initialization options as well as the following options:
        :keyword mask: the XNAT mask resource name
        :keyword registration_resource: the XNAT registration resource
            name
        :keyword registration_technique: the
            class:`qipipe.pipeline.registration.RegistrationWorkflow`
            technique
        :keyword recursive_registration: the
            class:`qipipe.pipeline.registration.RegistrationWorkflow`
            recursive flag
        :keyword modeling_resource: the modeling resource name
        :keyword modeling_technique: the
            class:`qipipe.pipeline.modeling.ModelingWorkflow` technique
        """
        super(QIPipelineWorkflow, self).__init__(
            project=project, logger=logger(__name__), **opts
        )

        reg_tech_opt = opts.get('registration_technique')
        if reg_tech_opt:
            self.registration_technique = reg_tech_opt.lower()
            """The registration technique."""
        elif 'register' in actions:
            raise PipelineError('The registration technique was not specified.')

        reg_rsc_opt = opts.get('registration_resource')
        if reg_rsc_opt:
            reg_rsc = reg_rsc_opt
        elif 'register' in actions:
            reg_rsc = registration.generate_resource_name(self.registration_technique)
            self.logger.debug("Generated registration resource name %s" %
                               reg_rsc)
        else:
            reg_rsc = None
        self.registration_resource = reg_rsc
        """The registration XNAT resource name."""

        mdl_tech_opt = opts.get('modeling_technique')
        if mdl_tech_opt:
            self.modeling_technique = mdl_tech_opt.lower()
            """The modeling technique name."""
        elif 'model' in actions:
            raise PipelineError('The modeling technique was not specified.')

        mdl_rsc_opt = opts.get('modeling_resource')
        if mdl_rsc_opt:
            mdl_rsc = mdl_rsc_opt
        elif 'model' in actions:
            mdl_rsc = modeling.generate_resource_name(self.modeling_technique)
        else:
            mdl_rsc = None
        self.modeling_resource = mdl_rsc
        """The modeling XNAT resource name."""

        self.workflow = self._create_workflow(actions, **opts)
        """
        The pipeline execution workflow. The execution workflow is executed
        by calling the :meth:`run_with_dicom_input` or
        :meth:`run_with_scan_download` method.
        """

    def run_with_dicom_input(self, actions, scan_input, dest=None):
        """
        :param actions: the workflow actions to perform
        :param scan_input: the :class:`qipipe.staging.iterator.ScanInput`
            object
        :param dest: the TCIA staging destination directory (default is
            the current working directory)
        """
        # Set the workflow input.
        input_spec = self.workflow.get_node('input_spec')
        input_spec.inputs.collection = scan_input.collection
        input_spec.inputs.subject = scan_input.subject
        input_spec.inputs.session = scan_input.session
        input_spec.inputs.scan = scan_input.scan
        input_spec.inputs.registered = []

        # If staging is enabled, then set the staging iterables.
        if 'stage' in actions:
            staging.set_workflow_iterables(self.workflow, scan_input, dest)
        # If roi is enabled and has input, then set the roi function inputs.
        if 'roi' in actions and scan_input.iterators.roi:
            self._set_roi_inputs(*scan_input.iterators.roi)
        # Execute the workflow.
        self.logger.debug("Running the pipeline on %s %s scan %d." %
                           (scan_input.subject, scan_input.session,
                            scan_input.scan))
        self._run_workflow(self.workflow)
        self.logger.debug("Completed pipeline execution on %s %s scan %d." %
                           (scan_input.subject, scan_input.session,
                            scan_input.scan))

    def run_with_scan_download(self, xnat, project, subject, session, scan,
                               actions, is_existing_registration_resource):
        """
        Runs the execution workflow on downloaded scan image files.

        :param project: the project name
        :param subject: the subject name
        :param session: the session name
        :param scan: the scan number
        :param actions: the workflow actions
        :param is_existing_registration_resource: flag indicating
            whether an existing registration resource name was
            specfied
        """
        self.logger.debug("Processing the %s %s %s Scan %d volumes..." %
                           (project, subject, session, scan))
        # Get the volume numbers.
        scan_obj = xnat.find_one(project, subject, session, scan=scan)
        if not scan_obj:
            raise NotFoundError("The pipeline did not find a %s %s %s"
                                " scan %d." %
                                (project, subject, session, scan))
        # The NIFTI resource contains the volume files.
        rsc_obj = scan_obj.resource('NIFTI')
        if not rsc_obj.exists():
            raise NotFoundError("The pipeline did not find a %s %s %s"
                                " scan %d NIFTI resource." %
                                (project, subject, session, scan))
        # The volume files.
        files = rsc_obj.files().get()
        if not files:
            raise ArgumentError("There are no pipeline %s %s %s Scan %d"
                                " NIFTI volumes" %
                                (project, subject, session, scan))

        # If the registration resource already exists in XNAT, then
        # partition the scan image file names into those which are
        # already registered and those which need to be registered.
        if is_existing_registration_resource:
            registered, unregistered = self._partition_registered(
                xnat, project, subject, session, scan, files
            )
        else:
            registered = []
            unregistered = files
        # Validate and log the partition.
        if 'register' in actions:
            if registered:
                if unregistered:
                    self.logger.debug("Skipping registration of %d"
                                       " previously registered %s %s %s"
                                       " Scan %d volumes:" %
                                       (len(registered), project, subject,
                                        session, scan))
                    self.logger.debug("%s" % registered)
                else:
                    self.logger.debug("Skipping %s %s %s Scan %d"
                                       " registration, since all volumes"
                                       " are already registered." %
                                       (project, subject, session, scan))
            if unregistered:
                self.logger.debug("Registering %d %s %s %s Scan %d"
                                   " volumes:" %
                                   (len(unregistered), project, subject,
                                    session, scan))
                self.logger.debug("%s" % unregistered)
        elif unregistered and is_existing_registration_resource:
            self.logger.error("The pipeline %s %s %s Scan %d register"
                                " action is not specified but there are"
                                "  %d unregistered volumes:" %
                                (project, subject, session, scan,
                                 len(unregistered)))
            self.logger.error("%s" % unregistered)
            raise ArgumentError("The pipeline %s %s %s Scan %d register"
                                " action is not specified but there are"
                                " unregistered volumes" %
                                (project, subject, session, scan))
        else:
            self.logger.debug("Processing %d %s %s %s Scan %d volumes:" %
                               (len(unregistered), project, subject, session,
                                scan))
            self.logger.debug("%s" % unregistered)

        # Set the workflow input.
        input_spec = self.workflow.get_node('input_spec')
        input_spec.inputs.subject = subject
        input_spec.inputs.session = session
        input_spec.inputs.scan = scan
        input_spec.inputs.registered = registered

        # Execute the workflow.
        self._run_workflow(self.workflow)
        self.logger.debug("Registered %d %s %s %s Scan %d volumes:" %
                           (len(unregistered), project, subject, session,
                            scan))
        self.logger.debug("%s" % unregistered)

    def _set_roi_inputs(self, *inputs):
        """
        :param collection: the image collection name
        :param subject: the subject name
        :param session: the session name
        :param scan: the scan number
        :param inputs: the :meth:`roi` inputs
        """
        # Set the roi function inputs.
        roi_node = self.workflow.get_node('roi')
        roi_node.inputs.in_rois = inputs
        roi_node.inputs.opts = dict(base_dir=self.base_dir)

    def _partition_registered(self, xnat, project, subject, session, scan,
                              files):
        """
        Partitions the given volume file names into those which have a
        corresponding registration resource file and those which don't.

        :return: the (registered, unregistered) file names
        """
        # The XNAT registration object.
        if not self.registration_resource:
            raise ArgumentError("The registration resource is missing")
        reg_obj = xnat.find_one(project, subject, session, scan=scan,
                                resource=self.registration_resource)
        if not reg_obj:
            raise ArgumentError("The registration resource %s does not exist"
                                " in XNAT: %s" % self.registration_resource)

        # The realigned files.
        registered = set(reg_obj.files().get())
        # The unregistered volume numbers.
        unregistered = set(files) - registered
        self.logger.debug("The %s %s %s resource has %d registered volumes"
                           " and %d unregistered volumes." %
                           (project, subject, session, len(registered),
                            len(unregistered)))

        return registered, unregistered

    def _create_workflow(self, actions, **opts):
        """
        Builds the reusable pipeline workflow described in
        :class:`qipipe.pipeline.qipipeline.QIPipeline`.

        :param actions: the actions to perform
        :param opts: the constituent workflow initializer options
        :return: the Nipype workflow
        """
        # This is a long method body with the following stages:
        #
        # 1. Gather the options.
        # 2. Create the constituent workflows.
        # 3. Tie together the constituent workflows.
        #
        # The constituent workflows are created in back-to-front order,
        # i.e. modeling, registration, mask, roi, staging.
        # This order makes it easier to determine whether to create
        # an upstream workflow depending on the presence of downstream
        # workflows, e.g. the mask is not created if registration
        # is not performed.
        #
        # By contrast, the workflows are tied together in front-to-back
        # order.
        self.logger.debug("Building the pipeline execution workflow"
                            " for the actions %s..." % actions)
        # The execution workflow.
        exec_wf = pe.Workflow(name='qipipeline', base_dir=self.base_dir)

        # The resource names.
        reg_rsc_opt = opts.get('registration_resource')
        recursive_reg_opt = opts.get('recursive_registration')
        mdl_rsc_opt = opts.get('modeling_resource')
        mask_rsc_opt = opts.get('mask')
        scan_ts_rsc_opt = opts.get('scan_time_series')
        reg_ts_rsc_opt = opts.get('realigned_time_series')

        # The technique options.
        reg_tech_opt = opts.get('registration_technique')
        mdl_tech_opt = opts.get('modeling_technique')

        if 'model' in actions:
            mdl_wf_gen = ModelingWorkflow(parent=self, technique=mdl_tech_opt,
                                          resource=self.modeling_resource)
            mdl_wf = mdl_wf_gen.workflow
            self.modeling_resource = mdl_wf_gen.resource
        else:
            mdl_wf = None

        # The registration workflow node.
        if 'register' in actions:
            reg_inputs = ['technique', 'project', 'subject', 'session',
                          'scan', 'resource', 'bolus_arrival_index', 
                          'in_files', 'mask', 'opts']
            # The registration function keyword options.
            if not self.registration_technique:
                raise ArgumentError('Missing the registration technique')
            if not self.registration_resource:
                raise ArgumentError('Missing the registration resource name')
            # Spell out the registration workflow options rather
            # than delegating to this qipipeline workflow as the
            # parent, since Nipype Function arguments must be
            # primitive.
            reg_opts = dict(
                base_dir=self.base_dir,
                distributable=self.is_distributable,
                dry_run=self.dry_run
            )
            if 'recursive_registration' in opts:
                reg_opts['recursive'] = opts['recursive_registration']
            # The registration function.
            reg_xfc = Function(input_names=reg_inputs,
                               output_names=['out_files'],
                               function=_register)
            reg_node = pe.Node(reg_xfc, name='register')
            reg_node.inputs.project = self.project
            reg_node.inputs.technique = self.registration_technique
            reg_node.inputs.resource = self.registration_resource
            reg_node.inputs.opts = reg_opts
            self.logger.info("Enabled registration.")
        else:
            self.logger.info("Skipping registration.")
            reg_node = None

        # The ROI workflow node.
        if 'roi' in actions:
            roi_inputs = ['project', 'subject', 'session', 'scan',
                          'time_series', 'in_rois', 'opts']
            roi_xfc = Function(input_names=roi_inputs, output_names=[],
                               function=roi)
            roi_node = pe.Node(roi_xfc, name='roi')
            roi_node.inputs.project = self.project
            roi_node.opts = dict(base_dir=self.base_dir,
                                 distributable=self.is_distributable)
            self.logger.info("Enabled ROI conversion.")
        else:
            roi_node = None
            self.logger.info("Skipping ROI conversion.")

        # Registration and modeling require a mask.
        if (reg_node or mdl_wf) and not mask_rsc_opt:
            mask_wf_gen = MaskWorkflow(parent=self)
            mask_wf = mask_wf_gen.workflow
            self.logger.info("Enabled scan mask creation.")
        else:
            mask_wf = None
            self.logger.info("Skipping scan mask creation.")

        # The staging workflow.
        if 'stage' in actions:
            stg_wf_gen = StagingWorkflow(parent=self)
            stg_wf = stg_wf_gen.workflow
            self.logger.info("Enabled staging.")
        else:
            stg_wf = None
            self.logger.info("Skipping staging.")

        # Validate that there is at least one constituent workflow.
        if not any([stg_wf, roi_node, reg_node, mdl_wf]):
            raise ArgumentError("No workflow was enabled.")

        # The workflow input fields.
        input_fields = ['subject', 'session', 'scan']
        iter_volume_fields = ['volume']
        # The staging workflow has additional input fields.
        # Partial registration requires the unregistered volumes input.
        if stg_wf:
            input_fields.append('collection')
            iter_volume_fields.append('dest')
        elif reg_node:
            input_fields.append('registered')

        # The workflow input node.
        input_spec_xfc = IdentityInterface(fields=input_fields)
        input_spec = pe.Node(input_spec_xfc, name='input_spec')
        # Staging, registration, and mask require a volume iterator node.
        # Modeling requires a volume iterator node if and only if the
        # following conditions hold:
        # * modeling is performed on the scan time series, and
        # * the scan time series is not specified
        model_reg = reg_node or reg_ts_rsc_opt
        model_scan = not model_reg
        model_vol = model_scan and not scan_ts_rsc_opt
        if stg_wf or reg_node or mask_wf or (mdl_wf and model_vol):
            iter_volume_xfc = IdentityInterface(fields=iter_volume_fields)
            iter_volume = pe.Node(iter_volume_xfc, name='iter_volume')
        # Staging requires a DICOM iterator node.
        if stg_wf:
            iter_dicom_xfc = IdentityInterface(fields=['volume', 'dicom_file'])
            iter_dicom = pe.Node(iter_dicom_xfc, name='iter_dicom')
            exec_wf.connect(iter_volume, 'volume', iter_dicom, 'volume')

        # Stitch together the workflows:

        # If staging is enabled, then stage the DICOM input.
        if stg_wf:
            for field in input_spec.inputs.copyable_trait_names():
                exec_wf.connect(input_spec, field,
                                stg_wf, 'input_spec.' + field)
            for field in iter_volume.inputs.copyable_trait_names():
                exec_wf.connect(iter_volume, field,
                                stg_wf, 'iter_volume.' + field)
            exec_wf.connect(iter_dicom, 'dicom_file',
                            stg_wf, 'iter_dicom.dicom_file')

        # Some workflows require the scan volumes, as follows:
        # * If staging is enabled, then collect the staged NIfTI
        #   scan images.
        # * Otherwise, if registration is enabled and there is not
        #   yet a scan time series, then download the staged XNAT
        #   scan images.
        # In either of the above cases, the staged images are collected
        # in a node named 'staged' with output 'images', which is used
        # later in the pipeline.
        # Otherwise, there is no staged node.
        staged = None
        if reg_node or roi_node or mask_wf or (mdl_wf and not model_reg):
            if stg_wf:
                staged = pe.JoinNode(IdentityInterface(fields=['out_files']),
                                    joinsource='iter_volume', name='staged')
                exec_wf.connect(stg_wf, 'output_spec.image',
                                staged, 'out_files')
            elif reg_node or not scan_ts_rsc_opt:
                dl_scan_xfc = XNATDownload(project=self.project,
                                           resource='NIFTI')
                staged = pe.Node(dl_scan_xfc, name='staged')
                exec_wf.connect(input_spec, 'subject', staged, 'subject')
                exec_wf.connect(input_spec, 'session', staged, 'session')
                exec_wf.connect(input_spec, 'scan', staged, 'scan')

        # All downstream actions require a scan time series.
        if reg_node or mask_wf or roi_node or mdl_wf:
            # If there is a scan time series, then download it.
            # Otherwise, if staging is enabled, then stack the resulting
            # staged 3D images into the scan time series.
            # Any other case is an error.
            if scan_ts_rsc_opt:
                dl_scan_ts_xfc = XNATDownload(project=self.project,
                                              resource=SCAN_TS_RSC)
                scan_ts = pe.Node(dl_scan_ts_xfc,
                                  name='download_scan_time_series')
                exec_wf.connect(input_spec, 'subject', scan_ts, 'subject')
                exec_wf.connect(input_spec, 'session', scan_ts, 'session')
                exec_wf.connect(input_spec, 'scan', scan_ts, 'scan')
            elif staged:
                # Merge the staged files.
                scan_ts_xfc = MergeNifti(out_format=SCAN_TS_RSC)
                scan_ts = pe.Node(scan_ts_xfc, name='merge_volumes')
                exec_wf.connect(staged, 'out_files', scan_ts, 'in_files')
                self.logger.debug('Connected staging to scan time series merge.')
                # Upload the time series.
                ul_scan_ts_xfc = XNATUpload(project=self.project,
                                            resource=SCAN_TS_RSC,
                                            modality='MR')
                ul_scan_ts = pe.Node(ul_scan_ts_xfc,
                                     name='upload_scan_time_series')
                exec_wf.connect(input_spec, 'subject', ul_scan_ts, 'subject')
                exec_wf.connect(input_spec, 'session', ul_scan_ts, 'session')
                exec_wf.connect(input_spec, 'scan', ul_scan_ts, 'scan')
                exec_wf.connect(scan_ts, 'out_file', ul_scan_ts, 'in_files')
            else:
                raise NotFoundError('The workflow requires a scan time series')

        # Registration and modeling require a mask and bolus arrival.
        if reg_node or mdl_wf:
            # If a mask resource name was specified, then download the mask.
            # Otherwise, make the mask.
            if mask_rsc_opt:
                dl_mask_xfc = XNATDownload(project=self.project,
                                           resource=mask_rsc_opt)
                download_mask = pe.Node(dl_mask_xfc, name='download_mask')
                exec_wf.connect(input_spec, 'subject', download_mask, 'subject')
                exec_wf.connect(input_spec, 'session', download_mask, 'session')
                exec_wf.connect(input_spec, 'scan', download_mask, 'scan')
            else:
                assert mask_wf, "The mask workflow is missing"
                exec_wf.connect(input_spec, 'subject',
                                mask_wf, 'input_spec.subject')
                exec_wf.connect(input_spec, 'session',
                                mask_wf, 'input_spec.session')
                exec_wf.connect(input_spec, 'scan',
                                mask_wf, 'input_spec.scan')
                exec_wf.connect(scan_ts, 'out_file',
                                mask_wf, 'input_spec.time_series')
                self.logger.debug('Connected scan time series to mask.')

            # Compute the bolus arrival from the scan time series.
            bolus_arv_xfc = Function(input_names=['time_series'],
                                     output_names=['bolus_arrival_index'],
                                     function=bolus_arrival_index_or_zero)
            bolus_arv = pe.Node(bolus_arv_xfc, name='bolus_arrival_index')
            exec_wf.connect(scan_ts, 'out_file', bolus_arv, 'time_series')
            self.logger.debug('Connected scan time series to bolus arrival.')

        # If ROI is enabled, then convert the ROIs.
        if roi_node:
            exec_wf.connect(input_spec, 'subject', roi_node, 'subject')
            exec_wf.connect(input_spec, 'session', roi_node, 'session')
            exec_wf.connect(input_spec, 'scan', roi_node, 'scan')
            exec_wf.connect(scan_ts, 'out_file', roi_node, 'time_series')
            self.logger.debug('Connected the scan time series to ROI.')

        # If registration is enabled, then register the unregistered
        # staged images.
        if reg_node:
            # There must be staged files.
            if not staged:
                raise NotFoundError('Registration requires scan input')
            exec_wf.connect(input_spec, 'subject', reg_node, 'subject')
            exec_wf.connect(input_spec, 'session', reg_node, 'session')
            exec_wf.connect(input_spec, 'scan', reg_node, 'scan')

            # If the registration input files were downloaded from
            # XNAT, then select only the unregistered files.
            if stg_wf:
                exec_wf.connect(staged, 'out_files', reg_node, 'in_files')
            else:
                exc_regd_xfc = Function(input_names=['in_files', 'exclusions'],
                                         output_names=['out_files'],
                                         function=exclude_files)
                exclude_registered = pe.Node(exc_regd_xfc,
                                             name='exclude_registered')
                exec_wf.connect(staged, 'out_files',
                                exclude_registered, 'in_files')
                exec_wf.connect(input_spec, 'registered',
                                exclude_registered, 'exclusions')
                exec_wf.connect(exclude_registered, 'out_files',
                                reg_node, 'in_files')
            self.logger.debug('Connected staging to registration.')

            # The mask input.
            if mask_wf:
                exec_wf.connect(mask_wf, 'output_spec.mask', reg_node, 'mask')
                self.logger.debug('Connected mask to registration.')
            else:
                exec_wf.connect(download_mask, 'out_file', reg_node, 'mask')
            # The bolus arrival.
            exec_wf.connect(bolus_arv, 'bolus_arrival_index',
                            reg_node, 'bolus_arrival_index')
            self.logger.debug('Connected bolus arrival to registration.')

        # If the modeling workflow is enabled, then model the scan or realigned
        # images.
        if mdl_wf:
            exec_wf.connect(input_spec, 'subject',
                            mdl_wf, 'input_spec.subject')
            exec_wf.connect(input_spec, 'session',
                            mdl_wf, 'input_spec.session')
            exec_wf.connect(input_spec, 'scan',
                            mdl_wf, 'input_spec.scan')
            # The modeling profile file path.
            mdl_profile_dest = os.path.join(self.base_dir, MODELING_PROFILE_FILE)
            mdl_input_spec = mdl_wf.get_node('input_spec')
            mdl_input_spec.inputs.profile_dest = mdl_profile_dest
            # The mask input.
            if mask_wf:
                exec_wf.connect(mask_wf, 'output_spec.mask',
                                mdl_wf, 'input_spec.mask')
                self.logger.debug('Connected mask to modeling.')
            else:
                exec_wf.connect(download_mask, 'out_file',
                                mdl_wf, 'input_spec.mask')
            # The bolus arrival.
            exec_wf.connect(bolus_arv, 'bolus_arrival_index',
                            mdl_wf, 'input_spec.bolus_arrival_index')
            self.logger.debug('Connected bolus arrival to modeling.')

            # If registration is enabled, then the registration 4D
            # time series is created by that workflow, otherwise
            # download the previously created time series.
            if reg_ts_rsc_opt:
                # Download the XNAT time series file.
                ts_dl_xfc = XNATDownload(project=self.project,
                                         resource=reg_ts_rsc_opt)
                reg_ts = pe.Node(ts_dl_xfc, name='download_reg_time_series')
                exec_wf.connect(input_spec, 'subject', reg_ts, 'subject')
                exec_wf.connect(input_spec, 'session', reg_ts, 'session')
                exec_wf.connect(input_spec, 'scan', reg_ts, 'scan')
                exec_wf.connect(reg_ts, 'out_file',
                                mdl_wf, 'input_spec.time_series')
            elif self.registration_resource:
                # Merge the realigned images to 4D.
                reg_ts_rsc = self.registration_resource + '_ts'
                merge_reg = pe.Node(MergeNifti(out_format=reg_ts_rsc),
                                    name='merge_reg')

                # If the registration resource name was specified,
                # then download the previously realigned images.
                if reg_rsc_opt:
                    reg_dl_xfc = XNATDownload(project=self.project,
                                              resource=reg_rsc_opt)
                    download_reg = pe.Node(reg_dl_xfc,
                                           name='download_realigned_images')
                    exec_wf.connect(input_spec, 'subject',
                                    download_reg, 'subject')
                    exec_wf.connect(input_spec, 'session',
                                    download_reg, 'session')
                    exec_wf.connect(input_spec, 'scan',
                                    download_reg, 'scan')
                    if reg_node:
                        # Merge the previously and newly realigned images.
                        concat_reg = pe.Node(Merge(2),
                                             name='concat_reg')
                        exec_wf.connect(download_reg, 'out_files',
                                        concat_reg, 'in1')
                        exec_wf.connect(reg_node, 'out_files',
                                        concat_reg, 'in2')
                        exec_wf.connect(concat_reg, 'out',
                                        merge_reg, 'in_files')
                    else:
                        # All of the realigned files were downloaded.
                        exec_wf.connect(download_reg, 'out_files',
                                        merge_reg, 'in_files')
                elif reg_node:
                    # All of the realigned files were created by the
                    # registration workflow.
                    exec_wf.connect(reg_node, 'out_files',
                                    merge_reg, 'in_files')
                else:
                    raise ArgumentError(
                        'The pipeline cannot perform modeling on the'
                        ' registration result, since the registration'
                        ' workflow is disabled and no registration resource'
                        ' was specified.')

                # Upload the realigned time series to XNAT.
                upload_reg_ts_xfc = XNATUpload(project=self.project,
                                               resource=reg_ts_rsc,
                                               modality='MR')
                upload_reg_ts = pe.Node(upload_reg_ts_xfc,
                                        name='upload_reg_time_series')
                exec_wf.connect(input_spec, 'subject',
                                upload_reg_ts, 'subject')
                exec_wf.connect(input_spec, 'session',
                                upload_reg_ts, 'session')
                exec_wf.connect(input_spec, 'scan',
                                upload_reg_ts, 'scan')
                exec_wf.connect(merge_reg, 'out_file',
                                upload_reg_ts, 'in_files')

                # Pass the realigned time series to modeling.
                exec_wf.connect(merge_reg, 'out_file',
                                mdl_wf, 'input_spec.time_series')
                self.logger.debug('Connected registration to modeling.')
            else:
                # Model the scan input.
                exec_wf.connect(scan_ts, 'out_file',
                                mdl_wf, 'input_spec.time_series')

        # Set the configured workflow node inputs and plug-in options.
        self._configure_nodes(exec_wf)

        self.logger.debug("Created the %s workflow." % exec_wf.name)
        # If debug is set, then diagram the workflow graph.
        if self.logger.level <= logging.DEBUG:
            self.depict_workflow(exec_wf)

        return exec_wf

    def _run_workflow(self, workflow):
        """
        Overrides the superclass method to build the workflow if the
        *dry_run* instance variable flag is set.

        :param workflow: the workflow to run
        """
        super(QIPipelineWorkflow, self)._run_workflow(workflow)
        if self.dry_run:
            # Make a dummy empty file for simulating called workflows.
            _, path = tempfile.mkstemp()
            try:
                # If registration is enabled, then simulate it.
                if self.workflow.get_node('register'):
                    opts = dict(base_dir=self.workflow.base_dir, dry_run=True)
                    register('Dummy', 'Dummy', 'Dummy', 1, 0, path, [path], opts)
                # If ROI is enabled, then simulate it.
                if self.workflow.get_node('roi'):
                    opts = dict(base_dir=self.workflow.base_dir, dry_run=True)
                    # A dummy (lesion, slice index, in_file) ROI input tuple.
                    inputs = [(1, 1, path)]
                    roi('Dummy', 'Dummy', 'Dummy', 1, path, inputs, opts)
            finally:
                os.remove(path)


def exclude_files(in_files, exclusions):
    """
    :param in_files: the input file paths
    :param exclusions: the file names to exclude
    :return: the filtered input file paths
    """
    import os

    # Make the exclusions a set.
    exclusions = set(exclusions)

    # Filter the input files.
    return [location for location in in_files
            if os.path.split(location)[1] not in exclusions]


def bolus_arrival_index_or_zero(time_series):
    """
    :param time_series: the 4D time series image
    :return: the bolus arrival index, or zero if the arrival
        cannot be calculated
    """
    from qipipe.helpers.bolus_arrival import (bolus_arrival_index,
                                              BolusArrivalError)

    # Determines the bolus uptake. If it could not be determined,
    # then the first series is taken to be the uptake.
    try:
        return bolus_arrival_index(time_series)
    except BolusArrivalError:
        return 0


def _register(technique, project, subject, session, scan, resource,
             bolus_arrival_index, mask, in_files, opts):
    """
    A stub for the :meth:`register` method.

    :Note: contrary to Python convention, the opts method parameter
      is a required dictionary rather than a keyword aggregate (i.e.,
      ``**opts``). The Nipype ``Function`` interface does not support
      method aggregates. Similarly, the in_files parameter is a
      required list rather than a splat argument (i.e., *in_files).
    """
    from qipipe.pipeline.qipipeline import register

    return register(technique, project, subject, session, scan, resource,
                    bolus_arrival_index, mask, *in_files, **opts)


def register(technique, project, subject, session, scan, resource,
             bolus_arrival_index, mask, *in_files, **opts):
    """
    Runs the registration workflow on the given session scan images.

    :param technique: the registration technique
    :param project: the project name
    :param subject: the subject name
    :param session: the session name
    :param scan: the scan number
    :param resource: the registration resource name
    :param bolus_arrival_index: the bolus uptake volume index
    :param mask: the required scan mask file
    :param in_files: the input session scan 3D NIfTI images
    :param opts: the :meth:`qipipe.pipeline.registration.run` keyword
        options
    :return: the realigned image file path array
    """

    # Note: There is always a mask and resource argument. The mask
    # file and resource name are either specified as an input or
    # built by the workflow. The mask and resource is optional in
    # the registration run function. Therefore, the registration
    # run keyword options include the mask and resource, as well
    # as any other options passed from the Function node.
    if not mask:
        raise ArgumentError("The register method is missing the mask")
    if not resource:
        raise ArgumentError("The register method is missing the"
                            " XNAT registration resource name")
    reg_opts = dict(mask=mask, resource=resource)
    reg_opts.update(opts)

    # The input scan files sorted by volume number.
    volumes = sorted(in_files, key=_extract_volume_number)
    # The files up to and including bolus arrival are not realigned.
    start = bolus_arrival_index + 1
    unrealigned = volumes[:start]
    # The bolus arrival is the initial fixed image.
    ref_0 = volumes[bolus_arrival_index]
    # The files to realign.
    reg_inputs = volumes[start:]

    # Register the files.
    realigned = registration.run(technique, project, subject, session, scan,
                                 ref_0, *reg_inputs, **reg_opts)
    # Upload the unrealigned images into the XNAT registration
    # resource.
    upload_unreg = XNATUpload(project=project, subject=subject,
                              session=session, scan=scan,
                              resource=resource, in_files=unrealigned,
                              modality='MR')
    upload_unreg.run()

    # Return the unregistered and registered result.
    return unrealigned + realigned


def _extract_volume_number(location):
    _, fname = os.path.split(location)
    match = VOLUME_FILE_PAT.match(fname)
    if not match:
        raise ArgumentError("The volume file name without directory does"
                            " not begin with volume<number>: %s" % fname)
    return int(match.group(1))
    

def roi(project, subject, session, scan, time_series, in_rois, opts):
    """
    Runs the ROI workflow on the given session scan images.

    :Note: see the :meth:`register` note.

    :param project: the project name
    :param subject: the subject name
    :param session: the session name
    :param scan: the scan number
    :param time_series: the scan 4D time series
    :param in_rois: the :meth:`qipipe.pipeline.roi.run` input tuples
    :param opts: the :meth:`qipipe.pipeline.roi.run` keyword options
    """
    from qipipe.pipeline import roi

    roi.run(project, subject, session, scan, time_series, *in_rois, **opts)



import os
import re
from datetime import datetime
import functools
import dicom
from qiutil.logging import logger
from qiutil import dates
from qidicom import (meta, writer)
from .sarcoma_config import sarcoma_location

DATE_FMT = '%Y%m%d'
"""The DICOM date format is YYYYMMDD."""


COMMENT_PREFIX = re.compile('^TTC \d+(\/.\d*)? sec')
"""OHSU - the Image Comments tag value prefix."""


def fix_dicom_headers(collection, subject, *in_files, **opts):
    """
    Fix the given input DICOM files as follows:

    - Replace the ``Patient ID`` value with the subject number, e.g.
        ``Sarcoma001``

    - Add the ``Body Part Examined`` tag

    - Anonymize the ``Patient's Birth Date`` tag

    - Standardize the file name

    OHSU - The ``Body Part Examined`` tag is set as follows:

    - If the collection is ``Sarcoma``, then the body part is the
        :meth:`qipipe.staging.sarcoma_config.sarcoma_location`.
    
    - Otherwise, the body part is the capitalized collection name, e.g.
        ``BREAST``.

    OHSU - Remove extraneous ``Image Comments`` tag value content
    which might contain PHI.
        
    The output file name is standardized as follows:

    - The file name is lower-case

    - The file extension is ``.dcm``

    - Each non-word character is replaced by an underscore
    
    :param collection: the collection name
    :param subject: the input subject name
    :param opts: the following keyword arguments:
    :keyword dest: the location in which to write the modified files
        (default is the current directory)
    :return: the files which were created
    :raise StagingError: if the collection is not supported
    """
    # Make the tag name => value dictionary.
    if collection == 'Sarcoma':
        site = sarcoma_location(subject)
    else:
        site = collection.upper()
    # The tag editor.
    editor = meta.Editor(PatientID=subject, BodyPartExamined=site,
                         PatientsBirthDate=_anonymize_date,
                         ImageComments=_scrub_comment)
    # The destination directory.
    dest = opts.get('dest', os.getcwd())
    # The destination file namer.
    file_namer = functools.partial(_dest_file_name, dest=dest)
    logger(__name__).debug("Replacing the following DICOM tags: %s" %
                           editor.edits.keys())
    # An array to collect the edited files.
    dcm_files = []
    # Edit the DICOM files.
    for ds in writer.edit(*in_files, dest=file_namer):
        editor.edit(ds)
        dcm_files.append(ds.filename)
    logger(__name__).debug("Changed the DICOM tag values.")

    # Return the output file names.
    return [file_namer(f) for f in dcm_files]


def _anonymize_date(date):
    """
    :param date: the input date string
    :return: the anonymized date
    :rtype: str
    """
    # The unanonymized PHI date.
    phi = datetime.strptime(date, DATE_FMT)
    # The anonymized date.
    anon = dates.anonymize(phi)
    
    # Return the anonymized date as a string.
    return anon.strftime(DATE_FMT)


def _scrub_comment(comment):
    """
    :param comment: the comment tag value
    :return: the scrubbed comment
    """
    match = COMMENT_PREFIX.match(comment)
    
    return match.group(0) if match else comment


def _dest_file_name(in_file, dest):
    """
    Standardizes the given input file name.
    
    :param in_file: the input file name
    :param dest: the destination directory
    :return: the target output file name 
    """
    _, fname = os.path.split(in_file)
    # Replace non-word characters.
    fname = re.sub('\W', '_', fname.lower())
    # Add a .dcm extension, if necessary.
    _, ext = os.path.splitext(fname)
    if not ext:
        fname = fname + '.dcm'
    
    return os.path.join(dest, fname)



"""
This module updates the qiprofile database imaging information
from a XNAT scan.
"""
from qiutil.file import splitexts
from qiprofile_rest_client.helpers import database
from qiprofile_rest_client.model.imaging import (Session, SessionDetail,
                                                 Modeling, ParameterResult)
from ..helpers.constants import (SUBJECT_FMT, SESSION_FMT)
from ..pipeline.modeling import MODELING_PROFILE_FILE
from ..pipeline.modeling import INFERRED_R1_0_OUTPUTS as OUTPUTS
from . import modeling


class ImagingError(Exception):
    pass


def update(subject, scan):
    """
    Updates the imaging content for the given qiprofile REST Subject
    from the given XNAT scan.

    :param subject: the target qiprofile Subject to update
    :param scan: the XNAT scan object
    """
    # The parent XNAT experiment.
    exp = scan.parent()
    # The XNAT experiment must have a date.
    date = exp.attrs.get('date')
    if not date:
        raise ImagingError(
            "The XNAT %s %s Subject %d %s experiment is missing"
            " the visit date" % (subject.project, subject.collection,
                         subject.number, exp.label)
        )
    # If there is a qiprofile session with the same date,
    # then complain.
    if any(sess.date == date for sess in subject.sessions):
        raise ImagingError(
            "qiprofile %s %s Subject %d session with visit date %s"
            " already exists" % (subject.project, subject.collection,
                                 subject.number, date)
        )
    # Make the qiprofile session database object.
    session = _create_session(exp)
    # Add the session to the subject encounters in date order.
    subject.add_encounter(session)
    # Save the session detail.
    session.detail.save()
    # Save the subject.
    subject.save()


def _create_session(xnat_exp):
    """
    Makes the qiprofile Session object from the XNAT scan.
    
    :param xnat_exp: the XNAT experiment object
    :return: the qiprofile session object
    """
    # Make the qiprofile scans.
    scans = [_create_scan(xnat_scan) for xnat_scan in xnat_exp.scans()]

    # The modeling resources begin with 'pk_'.
    xnat_mdl_rscs = (rsc for rsc in xnat_scan.resources()
                     if rsc.label().startswith('pk_'))
    modelings = [_create_modeling(rsc) for rsc in xnat_mdl_rscs]

    # The session detail database object to hold the scans.
    detail = SessionDetail(scans=scans)
    # Save the detail first, since it is not embedded and we need to
    # set the detail reference to make the session.
    detail.save()

    # The XNAT experiment date is the qiprofile session date.
    date = xnat_exp.attrs.get('date')

    # Return the new qiprofile Session object.
    return Session(date=date, modelings=modelings, detail=detail)


def _create_scan(xnat_scan):
    """
    Makes the qiprofile Session object from the XNAT scan.
    
    :param xnat_scan: the XNAT scan object
    :return: the qiprofile scan object
    """
    # The modeling resources begin with 'pk_'.
    xnat_mdl_rscs = (rsc for rsc in xnat_scan.resources()
                     if rsc.label().startswith('pk_'))
    modelings = [_create_modeling(rsc) for rsc in xnat_mdl_rscs]

    # The qiprofile scan to embed in the SessionDetail document.
    for scan in scan.scans():
        _update_scan(session, scan)
    # The session detail database object to hold the scans.
    detail = SessionDetail(scans=[scan])
    # Save the detail first, since it is not embedded and we need to
    # set the detail reference to make the session.
    detail.save()
    
    return Session(date=date, modelings=modelings,
                   tumor_extents=tumor_extents, detail=detail)


def _create_modeling(xnat_resource):
    """
    Creates the qiprofile Modeling object from the given XNAT
    resource object.

    :param xnat_resource: the XNAT modeling resource object
    """
    # The XNAT modeling files.
    xnat_files = xnat_resource.files()
    
    # The profile parameters.
    profile_finder = (xnat_file for xnat_file in xnat_files
                        if xnat_file.label() == MODELING_PROFILE_FILE)
    xnat_profile_file = next(profile_finder, None)
    if not xnat_profile_file:
        raise ImagingError("The XNAT modeling resource %s does not contain"
                           " the file %s" %
                           (xnat_resource.label(), MODELING_PROFILE_FILE))
    profile_location = xnat_profile_file.get()
    with open(profile_location) as profile_file:
        rows = csv.reader(profile_file)
        row_dict = {row[0]: row[1:] for row in rows}
        technique = row_dict.get('technique')
        if not technique:
            raise ImagingError("The XNAT modeling resource %s %s file does"
                               " not contain the technique entry" %
                               (xnat_resource.label(), MODELING_PROFILE_FILE))
        technique = row_dict.get('technique')
        if not technique:
            raise ImagingError("The XNAT modeling resource %s %s file does"
                               " not contain the technique entry" %
                               (xnat_resource.label(), MODELING_PROFILE_FILE))
    
    # The qiprofile ModelingProtocol.
    profile_dict = dict(technique=technique, r1_parameters=r1_params,
                        fastfit_parameter=fastfit_params)
    ModelingProtocol.objects.get(profile_dict)
    
    # The qiprofile modeling output files.
    possible_output_files = {output + '.nii.gz' for output in OUTPUTS}
    output_xnat_files = (xnat_file.label() for xnat_file in xnat_files
                         if xnat_file.label() in possible_output_files)

def _update_scan(session, xnat_scan):
    """
    Updates the scan content for the given qiprofile session
    database object from the given XNAT scan object.

    :param session: the target qiprofile Session object to update
    :param xnat_scan: the XNAT scan object
    """
    # TODO
    pass
